from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from PySide6.QtCore import QAbstractTableModel, QSize, Qt
from PySide6.QtWidgets import QAbstractItemView, QHBoxLayout, QHeaderView, QMenu, QPushButton, QTableView, QVBoxLayout

from .view import InstanceView

_l = logging.getLogger(name=__name__)

if TYPE_CHECKING:
    import PySide6

    from angrmanagement.data.instance import Instance
    from angrmanagement.data.signatures import Signature, SignatureManager
    from angrmanagement.ui.workspace import Workspace


class QSignatureTableModel(QAbstractTableModel):
    """
    Signature table model.
    """

    Headers = [
        "Type",
        "Name",
        "Architecture",
        "Platform",
        "Compiler",
        "OS Name",
        "Matches",
    ]
    COL_TYPE = 0
    COL_NAME = 1
    COL_ARCH = 2
    COL_PLATFORM = 3
    COL_COMPILER = 4
    COL_OS_NAME = 5
    COL_MATCHES = 6

    def __init__(self, signature_mgr: SignatureManager) -> None:
        super().__init__()
        self.signature_mgr = signature_mgr
        self.signature_mgr.signatures.am_subscribe(self._on_signatures_updated)

    def _on_signatures_updated(self, **kwargs) -> None:  # pylint:disable=unused-argument
        self.beginResetModel()
        self.endResetModel()

    def rowCount(self, parent: PySide6.QtCore.QModelIndex = ...) -> int:  # pylint:disable=unused-argument
        return len(self.signature_mgr.signatures)

    def columnCount(self, parent: PySide6.QtCore.QModelIndex = ...) -> int:  # pylint:disable=unused-argument
        return len(self.Headers)

    # pylint:disable=unused-argument
    def headerData(self, section: int, orientation: PySide6.QtCore.Qt.Orientation, role: int = ...) -> Any:
        if role != Qt.ItemDataRole.DisplayRole:
            return None
        if section < len(self.Headers):
            return self.Headers[section]
        return None

    def data(self, index: PySide6.QtCore.QModelIndex, role: int = ...) -> Any:
        if not index.isValid():
            return None
        row = index.row()
        if row >= len(self.signature_mgr.signatures):
            return None
        col = index.column()
        if role == Qt.ItemDataRole.DisplayRole:
            return self._get_column_text(self.signature_mgr.signatures[row], col)
        else:
            return None

    def _get_column_text(self, sig: Signature, column: int) -> str:
        if column == self.COL_TYPE:
            return sig.type_name
        elif column == self.COL_NAME:
            return sig.sig_name
        elif column == self.COL_ARCH:
            return sig.arch
        elif column == self.COL_PLATFORM:
            return sig.platform
        elif column == self.COL_COMPILER:
            return sig.compiler
        elif column == self.COL_OS_NAME:
            return sig.os_name
        elif column == self.COL_MATCHES:
            m = self.signature_mgr.get_match_count(sig)
            return str(m) if m is not None else "N/A"
        else:
            raise AssertionError


class QSignatureTableWidget(QTableView):
    """
    Signature table widget.
    """

    def __init__(self, signature_mgr, workspace: Workspace) -> None:
        super().__init__()
        self.workspace = workspace
        self.signature_mgr = signature_mgr

        hheader = self.horizontalHeader()
        hheader.setVisible(True)

        vheader = self.verticalHeader()
        vheader.setVisible(False)
        vheader.setDefaultSectionSize(20)

        self.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.setHorizontalScrollMode(QAbstractItemView.ScrollMode.ScrollPerPixel)

        self.model: QSignatureTableModel = QSignatureTableModel(self.signature_mgr)
        self.setModel(self.model)

        for col in range(len(QSignatureTableModel.Headers)):
            hheader.setSectionResizeMode(col, QHeaderView.ResizeMode.ResizeToContents)
        hheader.setStretchLastSection(True)
        self.doubleClicked.connect(self._on_cell_double_click)

    #
    # Events
    #

    def closeEvent(self, event) -> None:
        self.model.shutdown()
        super().closeEvent(event)

    def contextMenuEvent(self, event) -> None:
        selected_rows = {i.row() for i in self.selectedIndexes()}
        sigs = [self.signature_mgr.signatures[r] for r in selected_rows]
        menu = QMenu("", self)
        if len(sigs):
            menu.addAction(
                "&Try applying signature(s)", lambda: self.signature_mgr.apply_signatures(sigs, dry_run=True)
            )
            menu.addAction("&Apply signature(s)", lambda: self.signature_mgr.apply_signatures(sigs, dry_run=False))
            # Add "View Matches" option if there's exactly one signature selected and it has matches
            if len(sigs) == 1:
                sig = sigs[0]
                matches = self.signature_mgr.get_matches(sig)
                if matches:
                    menu.addAction("&View Matches...", lambda: self._show_matches_dialog(sig, matches))
            menu.addSeparator()
        menu.addAction("&Load signature files...", self.signature_mgr.load_signatures)
        menu.exec_(event.globalPos())

    # pylint:disable=unused-argument
    def _on_cell_double_click(self, index) -> None:
        """Double-click: try applying signature if not yet applied, otherwise show matches dialog."""
        row = index.row()
        if row >= len(self.signature_mgr.signatures):
            return
        sig = self.signature_mgr.signatures[row]
        matches = self.signature_mgr.get_matches(sig)
        if matches:
            # Already has matches, show the dialog
            self._show_matches_dialog(sig, matches)
        else:
            # No matches yet, try applying the signature
            self.signature_mgr.apply_signatures([sig], dry_run=True)

    def _show_matches_dialog(self, sig, matches) -> None:
        """Show dialog displaying matched functions for a signature."""
        from angrmanagement.ui.dialogs.signature_matches import SignatureMatchesDialog

        dialog = SignatureMatchesDialog(sig, matches, self.workspace.main_instance, self.signature_mgr, self)
        dialog.exec()


class SignaturesView(InstanceView):
    """
    Signatures view that displays all loaded (FLIRT) signatures and their statuses.
    """

    def __init__(self, workspace: Workspace, default_docking_position: str, instance: Instance) -> None:
        super().__init__("signatures", workspace, default_docking_position, instance)
        self.base_caption = "Function Signatures"
        self._tbl_widget: QSignatureTableWidget | None = None
        self._init_widgets()
        self.reload()

    def reload(self) -> None:
        self.instance.signature_mgr.sync_from_angr()

    # pylint:disable=no-self-use
    def minimumSizeHint(self):
        return QSize(200, 200)

    def _init_widgets(self) -> None:
        # Header with buttons
        header_layout = QHBoxLayout()

        load_btn = QPushButton("Load Signature Files...")
        load_btn.clicked.connect(self.instance.signature_mgr.load_signatures)

        try_all_btn = QPushButton("Try Applying All Signatures")
        try_all_btn.clicked.connect(self._try_apply_all_signatures)

        bindiff_btn = QPushButton("Find Matches Via Bindiff")
        bindiff_btn.clicked.connect(self._find_matches_via_bindiff)

        header_layout.addWidget(load_btn)
        header_layout.addWidget(try_all_btn)
        header_layout.addWidget(bindiff_btn)
        header_layout.addStretch()

        # Main layout
        vlayout = QVBoxLayout()
        vlayout.addLayout(header_layout)
        self._tbl_widget = QSignatureTableWidget(self.instance.signature_mgr, self.workspace)
        vlayout.addWidget(self._tbl_widget)
        self.setLayout(vlayout)

    def _try_apply_all_signatures(self) -> None:
        """Try applying all signatures in the view."""
        all_sigs = list(self.instance.signature_mgr.signatures)
        if all_sigs:
            self.instance.signature_mgr.apply_signatures(all_sigs, dry_run=True)

    def _on_bindiff_matches_finished(self, matches_per_file: dict[str, dict[int, str]]) -> None:
        for filename, matches in matches_per_file.items():
            self.instance.signature_mgr.add_precomputed_signature("Bindif", filename, matches)

    def _find_matches_via_bindiff(self) -> None:
        """Launch a background job to find matches via bindiff."""
        from angrmanagement.data.jobs.bindiff_matching import BindiffMatchingJob

        # Show the file dialog to select binaries
        filenames = self.instance.signature_mgr.select_bindiff_base_files()
        if filenames:
            # Create and launch the background job
            job = BindiffMatchingJob(self.instance, filenames, on_results=self._on_bindiff_matches_finished)
            self.workspace.job_manager.add_job(job)



from __future__ import annotations

import datetime
import logging
import time
from enum import Enum
from functools import lru_cache
from typing import TYPE_CHECKING, Any

from angrmanagement.logic import GlobalInfo

if TYPE_CHECKING:
    from collections.abc import Callable

    from IPython.extensions.autoreload import ModuleReloader

    from angrmanagement.data.instance import Instance
    from angrmanagement.logic.jobmanager import JobContext


log = logging.getLogger(__name__)


@lru_cache(maxsize=1)
def _module_reloader() -> ModuleReloader:
    try:
        from IPython.extensions.autoreload import ModuleReloader  # pylint:disable=import-outside-toplevel

        m = ModuleReloader()
        m.enabled = True
        m.check_all = True
        m.check()
    except ImportError:
        m = None

    return m


class JobState(Enum):
    """
    The state of a job.
    """

    PENDING = 0
    RUNNING = 1
    FINISHED = 2
    CANCELLED = 3
    FAILED = 4


class Job:
    """
    The base class of all Jobs in angr management.
    """

    name: str
    progress_percentage: float
    start_at: float
    blocking: bool
    state: JobState = JobState.PENDING
    _on_finish: Callable[[Any], None] | None
    result: Any

    def __init__(self, name: str, on_finish: Callable[[Any], None] | None = None, blocking: bool = False) -> None:
        self.name = name
        self.progress_percentage = 0.0
        self.start_at = 0.0
        self.blocking = blocking
        self.cancelled = False
        self.result = None

        # callbacks
        self._on_finish = on_finish

        if GlobalInfo.autoreload:
            m = _module_reloader()
            if m is not None:
                prestate = dict(m.modules_mtimes)
                m.check()
                poststate = dict(m.modules_mtimes)
                if prestate != poststate:
                    log.warning("Auto-reload found changed modules")

    @property
    def time_elapsed(self) -> str:
        return str(datetime.timedelta(seconds=int(time.time() - self.start_at)))

    def start(self, ctx: JobContext):
        self.result = self.run(ctx)

    def run(self, ctx: JobContext):
        """Run the job. This method is called in a worker thread."""
        raise NotImplementedError

    def finish(self) -> None:
        """Runs after the job has finished in the GUI thread."""
        if self._on_finish is not None:
            self._on_finish(self.result)


class InstanceJob(Job):
    """
    A job that operates on an instance.
    """

    instance: Instance

    def __init__(
        self, name: str, instance: Instance, on_finish: Callable[[Any], None] | None = None, blocking: bool = False
    ):
        super().__init__(name, on_finish, blocking)
        self.instance = instance



from __future__ import annotations

from typing import TYPE_CHECKING

from angr.sim_type import SimStruct, SimTypeBottom, SimUnion, TypeRef
from PySide6.QtCore import QSize, Qt
from PySide6.QtGui import QBrush, QColor, QPainter
from PySide6.QtWidgets import QMessageBox, QSizePolicy, QWidget

from angrmanagement.config import Conf
from angrmanagement.ui.dialogs.type_editor import CTypeEditor, edit_field

if TYPE_CHECKING:
    from angr.knowledge_plugins.types import TypesStore

    from angrmanagement.ui import views

LINE_HEIGHT = 20
COL_WIDTH = 8


class QCTypeDef(QWidget):
    """
    A widget to display a C SimType.
    """

    def __init__(self, parent, ty: TypeRef, all_types: TypesStore, editor: views.TypesView | None = None) -> None:
        super().__init__(parent)

        self.type = ty
        self.text = ""  # this will be used for full-text search
        self.lines = [""]
        self.highlight = None  # which line should be highlighted
        self.all_types = all_types
        self.editor = editor

        self.setAttribute(Qt.WidgetAttribute.WA_Hover)
        self.setMouseTracking(True)

        self.refresh()

    def refresh(self):
        if self.type._arch is None:
            raise TypeError("Must provide SimTypes with arches to QTypeDef")

        self.text = f"typedef {self.type.type.c_repr(name=self.type.name, full=1)};"
        self.lines = self.text.split("\n")
        fields = None
        offsets = None
        for i, line in enumerate(self.lines):
            type_size = self.type.size if not isinstance(self.type.type, SimTypeBottom) else 0
            if i == 0:
                prefix = f"{type_size // self.type._arch.byte_width:08x}"
            elif isinstance(self.type.type, SimUnion):
                prefix = "00000000"
            elif isinstance(self.type.type, SimStruct):
                fieldno = i - 1
                if fields is None:
                    fields = list(self.type.type.fields)
                    offsets = self.type.type.offsets
                if fieldno < len(fields):
                    prefix = f"{offsets[fields[fieldno]]:08x}"
                else:
                    prefix = f"{type_size // self.type._arch.byte_width:08x}"
            else:
                raise TypeError(f"I don't know why a {type(self.type.type)} renders with more than one line")

            self.lines[i] = f"{prefix}  {line}"

        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        height = len(self.lines) * LINE_HEIGHT + LINE_HEIGHT // 2
        self.setMinimumHeight(height)
        self.setMaximumHeight(height)
        self.setMinimumWidth(max(len(line) for line in self.lines) * COL_WIDTH)

        self.repaint()

    def sizeHint(self):
        return QSize(self.minimumWidth(), self.minimumHeight())

    def leaveEvent(self, event) -> None:  # pylint: disable=unused-argument
        self.highlight = None

    def mouseMoveEvent(self, event) -> None:
        old_highlight = self.highlight
        self.highlight = min(max((event.pos().y() - 5) // LINE_HEIGHT, 0), len(self.lines) - 1)
        if old_highlight != self.highlight:
            self.repaint()

    def paintEvent(self, event) -> None:  # pylint: disable=unused-argument
        painter = QPainter(self)

        if self.highlight is not None:
            # TODO use config colors
            painter.fillRect(0, 5 + 20 * self.highlight, self.width(), 20, QBrush(QColor(0xC0, 0xC0, 0xC0, 0xFF)))

        painter.setFont(Conf.disasm_font)
        y = 20
        for line in self.lines:
            painter.drawText(0, y, line)
            y += 20

    def mouseDoubleClickEvent(self, event) -> None:  # pylint: disable=unused-argument
        if self.highlight is None:
            return

        if self.highlight != 0:
            fieldno = self.highlight - 1
            try:
                edited = edit_field(self.type.type, fieldno)
            except IndexError:
                pass
            else:
                if edited:
                    if self.editor is not None:
                        self.editor.reload()
                    else:
                        self.refresh()
                return

        dialog = CTypeEditor(
            None,
            self.type._arch,
            self.text,
            multiline=True,
            editing_single=self.type.name,
            predefined_types=self.all_types,
        )
        dialog.exec_()
        if dialog.main_result:
            name, ty = dialog.main_result[0]
            if name is not None and stript(name) != stript(self.type.name):
                if name in self.all_types:
                    QMessageBox.warning(None, "Duplicate type name", f"The name {name} is already taken.")
                else:
                    # TODO handle the struct/union aliases properly
                    self.all_types.rename(self.type.name, stript(name))
            self.type.type = ty
            if self.editor is not None:
                self.editor.reload()
            else:
                self.refresh()


def stript(s):
    return s.removeprefix("struct ").removeprefix("union ")



from __future__ import annotations

import logging
import multiprocessing
import subprocess
import tempfile
from concurrent.futures import ProcessPoolExecutor, as_completed
from pathlib import Path
from typing import TYPE_CHECKING

import angr

if TYPE_CHECKING:
    from angrmanagement.data.instance import Instance
    from angrmanagement.logic.jobmanager import JobContext

from .job import InstanceJob

_l = logging.getLogger(name=__name__)


def _process_binary(
    target_project: angr.Project, binary_path: str, arch_name: str, instruction_endness: str
) -> tuple[str, bool, str | None, dict]:
    """
    Process a single binary file for bindiff analysis.
    Returns tuple of (binary_path, success, error_message, matches)
    This function runs in a separate process.
    """
    try:
        main_opts = {"arch": arch_name, "endness": instruction_endness}
        base_project = angr.Project(binary_path, main_opts=main_opts)

        base_project.analyses.CFG(
            show_progressbar=False,
            normalize=True,
            resolve_indirect_jumps=True,
            detect_tail_calls=True,
        )

        bindiff = target_project.analyses.BinDiff(
            base_project,
            cfg_a=target_project.kb.cfgs.get_most_accurate(),
            cfg_b=base_project.kb.cfgs.get_most_accurate(),
        )

        out = {}
        for a0, a1 in bindiff.function_matches:
            # func_0 = target_project.kb.functions.get_by_addr(a0)
            func_1 = base_project.kb.functions.get_by_addr(a1)
            if (
                func_1.is_default_name
                or func_1.name.startswith("sub_")
                or "unknown" in func_1.name.lower()
                or "unresolve" in func_1.name.lower()
            ):
                continue
            out[a0] = func_1.name
        return (binary_path, True, None, out)

    except angr.errors.AngrCFGError as e:
        # Ignore empty regions as some .o files may contain no code
        if "Regions are empty" in str(e):
            return (binary_path, True, None, {})
        return (binary_path, False, str(e), None)
    except Exception as e:
        return (binary_path, False, str(e), None)


class BindiffMatchingJob(InstanceJob):
    """
    Describes a job for finding matches via bindiff analysis using multiprocessing.
    """

    def __init__(self, instance: Instance, filenames: list[str], on_finish=None, on_results=None) -> None:
        super().__init__("Finding matches via bindiff", instance, on_finish=on_finish)
        self.filenames = filenames
        self.on_results = on_results

    def run(self, ctx: JobContext) -> None:
        binaries_to_process = []
        temp_dirs = []

        matches_per_file = {}

        for filename in self.filenames:
            if filename.endswith(".a"):
                tmpdir = tempfile.mkdtemp()
                temp_dirs.append(tmpdir)
                try:
                    subprocess.check_call(
                        ["ar", "x", Path(filename).absolute()],
                        cwd=tmpdir,
                    )
                    tmpdir_path = Path(tmpdir)
                    object_files = list(tmpdir_path.glob("*.o"))
                    binaries_to_process.extend([(filename, str(f)) for f in object_files])
                    matches_per_file[filename] = {}
                    _l.info("Extracted %d object files from %s", len(object_files), filename)
                except Exception:
                    _l.warning("Failed to unarchive %s using ar command. Is ar installed?", filename)
            else:
                binaries_to_process.append((filename, filename))
                matches_per_file[filename] = {}

        if not binaries_to_process:
            ctx.set_progress(1.0, "No binaries to process")
            return

        arch_name = self.instance.project.arch.name
        instruction_endness = self.instance.project.arch.instruction_endness

        max_workers = min(max(multiprocessing.cpu_count() - 1, 1), len(binaries_to_process))
        _l.info("Processing %d binaries with %d workers", len(binaries_to_process), max_workers)

        total_binaries = len(binaries_to_process)
        completed = 0
        failed = 0

        num_found_matches = 0
        num_conflicts = 0

        try:
            # Process binaries in parallel
            with ProcessPoolExecutor(max_workers=max_workers) as executor:
                future_to_binary = {
                    executor.submit(
                        _process_binary, self.instance.project.am_obj, binary, arch_name, instruction_endness
                    ): (original_file, binary)
                    for original_file, binary in binaries_to_process
                }

                for future in as_completed(future_to_binary):
                    original_file, binary = future_to_binary[future]
                    try:
                        binary_path, success, error_msg, output = future.result()
                        if success:
                            completed += 1
                            _l.debug("Successfully processed %s", Path(binary_path).name)
                            if output:
                                existing_matches = matches_per_file[original_file]
                                for addr, name in output.items():
                                    if addr in existing_matches:
                                        num_conflicts += 1
                                    else:
                                        existing_matches[addr] = name
                                        num_found_matches += 1
                        else:
                            failed += 1
                            _l.warning("Failed to process %s: %s", Path(binary_path).name, error_msg)
                    except Exception as e:
                        failed += 1
                        _l.exception("Exception processing %s: %s", Path(binary).name, e)

                    progress = (completed + failed) / total_binaries * 100.0
                    ctx.set_progress(
                        progress,
                        f"{completed + failed}/{total_binaries} binaries ({failed} failed); "
                        f"Found {num_found_matches} matched functions; {num_conflicts} conflicts",
                    )

        finally:
            # Clean up temporary directories
            import shutil

            for tmpdir in temp_dirs:
                try:
                    shutil.rmtree(tmpdir)
                except Exception as e:
                    _l.warning("Failed to clean up temporary directory %s: %s", tmpdir, e)

        ctx.set_progress(1.0, f"Bindiff matching complete: {completed} succeeded, {failed} failed")
        _l.info("Bindiff matching complete: %d succeeded, %d failed", completed, failed)

        # Run the on_finish callback
        if self.on_results:
            self.on_results(matches_per_file)

    def __repr__(self) -> str:
        return "BindiffMatchingJob"



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import pyvex

try:
    import pypcode
except ImportError:
    pypcode = None

from PySide6.QtGui import QTextCharFormat

from angrmanagement.config import Conf

from .base_objects import BlockTreeNode

if TYPE_CHECKING:
    from archinfo import RegisterOffset, TmpVar


OBJ_CLASS_TO_QBLOCKCODE_CLASS = {}


class QIROpObj(BlockTreeNode):
    """
    Renders a Lifter IR object
    """

    @staticmethod
    def fmt() -> QTextCharFormat:
        fmt = QTextCharFormat()
        fmt.setForeground(Conf.disasm_view_ir_default_color)
        return fmt

    def __init__(self, obj: Any, *args, irobj=None, **kwargs) -> None:
        self.irobj = irobj or obj
        super().__init__(obj, *args, **kwargs)

    def create_subobjs(self, obj) -> None:
        self.add_irobj(obj.obj)

    def add_irobj(self, obj) -> None:
        subobjcls = OBJ_CLASS_TO_QBLOCKCODE_CLASS.get(type(obj), QIROpTextObj)
        subobj = subobjcls(obj, self.infodock, parent=self, options=self.options, irobj=self.irobj)
        self._add_subobj(subobj)


class QIROpTextObj(QIROpObj):
    """
    Renders a Lifter IR object using the object's __str__, or as hexadecimal
    if an integer type.
    """

    def create_subobjs(self, obj: Any) -> None:
        if isinstance(obj, int):
            self.add_text(f"{obj:#x}")
        else:
            self.add_text(str(obj))


class QIrOpPcodeOp(QIROpTextObj):
    """
    Renders a P-code op.
    """

    def create_subobjs(self, obj: pypcode.PcodeOp) -> None:  # type: ignore[reportInvalidTypeForm]
        self.add_text(pypcode.PcodePrettyPrinter.fmt_op(obj))  # type: ignore


if pypcode:
    OBJ_CLASS_TO_QBLOCKCODE_CLASS.update({pypcode.PcodeOp: QIrOpPcodeOp})


class QIROpVexConstObj(QIROpTextObj):
    """
    Renders a pyvex.expr.Const
    """

    @staticmethod
    def fmt() -> QTextCharFormat:
        fmt = QTextCharFormat()
        fmt.setForeground(Conf.disasm_view_operand_constant_color)
        return fmt

    def should_highlight(self) -> bool:
        return (
            isinstance(self.infodock.selected_block_tree_node, QIROpVexConstObj)
            and self.infodock.selected_block_tree_node.obj._con == self.obj._con
        )


class VexIRTmpWrapper:
    """
    A wrapper class for VEX temps
    """

    __slots__ = (
        "tid",
        "reg_name",
    )

    tid: TmpVar
    reg_name: str | None

    def __init__(self, tid: TmpVar, reg_name: str | None = None) -> None:
        self.tid = tid
        self.reg_name = reg_name or (f"t{self.tid}")

    def __str__(self) -> str:
        return self.reg_name if self.reg_name is not None else "None"


class VexIRRegWrapper:
    """
    A wrapper class for VEX registers
    """

    __slots__ = (
        "offset",
        "reg_name",
    )

    offset: RegisterOffset
    reg_name: str | None

    def __init__(self, offset: RegisterOffset, reg_name: str | None = None) -> None:
        self.offset = offset
        self.reg_name = reg_name or (f"offset={self.offset}")

    def __str__(self) -> str:
        return self.reg_name if self.reg_name is not None else "None"


class QIROpVexWrTmpObj(QIROpTextObj):
    """
    Renders a pyvex.stmt.WrTmp
    """

    def create_subobjs(self, obj: pyvex.stmt.WrTmp) -> None:
        irsb = self.irobj.irsb
        self.add_irobj(VexIRTmpWrapper(obj.tmp))
        self.add_text(" = ")
        if isinstance(obj.data, pyvex.expr.Get):
            reg_name = irsb.arch.translate_register_name(obj.data.offset, obj.data.result_size(irsb.tyenv) // 8)
            self.add_irobj(VexIRRegWrapper(obj.data.offset, reg_name))  # type: ignore
        else:
            self.add_irobj(obj.data)


class QIROpVexRdTmpObj(QIROpTextObj):
    """
    Renders a pyvex.expr.RdTmp
    """

    def create_subobjs(self, obj: pyvex.expr.RdTmp) -> None:
        self.add_irobj(VexIRTmpWrapper(obj.tmp))


class QIROpVexTmpObj(QIROpTextObj):
    """
    Renders a VEX temporary
    """

    @staticmethod
    def fmt() -> QTextCharFormat:
        fmt = QTextCharFormat()
        fmt.setForeground(Conf.disasm_view_operand_color)
        return fmt

    def should_highlight(self) -> bool:
        return (
            isinstance(self.infodock.selected_block_tree_node, QIROpVexTmpObj)
            and self.infodock.selected_block_tree_node.obj.tid == self.obj.tid
        )


class QIROpVexRegObj(QIROpTextObj):
    """
    Renders a VEX register
    """

    @staticmethod
    def fmt() -> QTextCharFormat:
        fmt = QTextCharFormat()
        fmt.setForeground(Conf.disasm_view_operand_color)
        return fmt

    def should_highlight(self) -> bool:
        return (
            isinstance(self.infodock.selected_block_tree_node, QIROpVexRegObj)
            and self.infodock.selected_block_tree_node.obj.offset == self.obj.offset
        )


class QIROpVexStoreObj(QIROpTextObj):
    """
    Renders a pyvex.stmt.Store
    """

    def create_subobjs(self, obj: pyvex.stmt.Store) -> None:
        # "ST%s(%s) = %s" % (self.endness[-2:].lower(), self.addr, self.data)
        self.add_text(f"ST{obj.endness[-2:].lower()}(")
        self.add_irobj(obj.addr)
        self.add_text(") = ")
        self.add_irobj(obj.data)


class QIROpVexLoadObj(QIROpTextObj):
    """
    Renders a pyvex.expr.Load
    """

    def create_subobjs(self, obj: pyvex.expr.Load) -> None:
        self.add_text(f"LD{obj.end[-2:].lower()}:{obj.ty[4:]}(")
        self.add_irobj(obj.addr)
        self.add_text(")")


class QIROpVexPutObj(QIROpTextObj):
    """
    Renders a pyvex.stmt.Put
    """

    def create_subobjs(self, obj: pyvex.stmt.Put) -> None:
        irsb = self.irobj.irsb
        reg_name = irsb.arch.translate_register_name(obj.offset, obj.data.result_size(irsb.tyenv) // 8)
        self.add_text("PUT(")
        self.add_irobj(VexIRRegWrapper(obj.offset, reg_name))  # type: ignore
        self.add_text(") = ")
        self.add_irobj(obj.data)


class QIROpVexExitObj(QIROpTextObj):
    """
    Renders a pyvex.stmt.Exit
    """

    def create_subobjs(self, obj: pyvex.stmt.Exit) -> None:
        irsb = self.irobj.irsb
        arch = irsb.arch
        reg_name = arch.translate_register_name(irsb.offsIP, arch.bits // 8)
        self.add_text("if (")
        self.add_irobj(obj.guard)
        self.add_text(") { PUT(")
        self.add_irobj(VexIRRegWrapper(obj.offsIP, reg_name))  # type: ignore
        self.add_text(") = ")
        self.add_irobj(obj.dst.value)
        self.add_text("; ")
        self.add_irobj(obj.jumpkind)
        self.add_text(" }")


class QIROpVexBinopObj(QIROpTextObj):
    """
    Renders a pyvex.expr.Binop
    """

    def create_subobjs(self, obj: pyvex.expr.Binop) -> None:
        self.add_text(obj.op[4:])
        self.add_text("(")
        self.add_irobj(obj.args[0])
        self.add_text(",")
        self.add_irobj(obj.args[1])
        self.add_text(")")


class QIROpVexUnopObj(QIROpTextObj):
    """
    Renders a pyvex.expr.Unop
    """

    def create_subobjs(self, obj: pyvex.expr.Unop) -> None:
        self.add_text(obj.op[4:])
        self.add_text("(")
        self.add_irobj(obj.args[0])
        self.add_text(")")


OBJ_CLASS_TO_QBLOCKCODE_CLASS.update(
    {
        pyvex.stmt.WrTmp: QIROpVexWrTmpObj,
        pyvex.expr.RdTmp: QIROpVexRdTmpObj,
        pyvex.stmt.Store: QIROpVexStoreObj,
        pyvex.expr.Load: QIROpVexLoadObj,
        pyvex.stmt.Put: QIROpVexPutObj,
        pyvex.stmt.Exit: QIROpVexExitObj,
        pyvex.expr.Const: QIROpVexConstObj,
        pyvex.expr.Binop: QIROpVexBinopObj,
        pyvex.expr.Unop: QIROpVexUnopObj,
        VexIRTmpWrapper: QIROpVexTmpObj,
        VexIRRegWrapper: QIROpVexRegObj,
    }
)



from __future__ import annotations

from collections import OrderedDict

import pycparser.c_parser
from angr import sim_type
from PySide6.QtCore import Qt
from PySide6.QtWidgets import QDialog, QDialogButtonBox, QLineEdit, QMessageBox, QPushButton, QVBoxLayout

from angrmanagement.config import Conf

from .set_comment import QCommentTextBox

# compatibility for pycparser 2.18 and 3.0
PycParseErrorCls = ()
try:
    import pycparser.plyparser

    PycParseErrorCls += (pycparser.plyparser.ParseError,)
except ImportError:
    pass

PycParseErrorCls += (pycparser.c_parser.ParseError,)


class CTypeEditor(QDialog):
    """
    A dialog to edit C types.

    :param parent:              The parent widget
    :param base_text:           The text to initially display in the editor
    :param multiline:           Whether the editor should be a multiline editor
    :param editing_single:      If set, the name of the single type we are editing

    :ivar list[(Optional[str], SimType)] main_result:
                                The entered types. Will be empty if the dialog was cancelled.
                                Will have only a single entry if editing_single is non-null.
    :ivar list[(Optional[str], SimType)] side_result:
                                Any additional types which were defined along the way.
    """

    def __init__(
        self,
        parent,
        arch,
        base_text: str = "",
        multiline: bool = False,
        editing_single: str | None = None,
        predefined_types=None,
    ) -> None:
        super().__init__(parent)

        self._editing_single = editing_single
        self.arch = arch
        self._predefined_types = predefined_types

        self.text = lambda: ""
        self.setText = lambda x: None
        self._ok_button: QPushButton | None
        self._init_widgets(base_text, multiline)

        self.setWindowTitle("Type editor")
        self.setWindowFlags(self.windowFlags() & ~Qt.WindowType.WindowContextHelpButtonHint)

        self.main_result: list[tuple[str, sim_type.SimType]] = []
        self.side_result: list[tuple[str, sim_type.SimType]] = []

    def _init_widgets(self, base_text, multiline) -> None:
        buttons = QDialogButtonBox(parent=self)
        buttons.setStandardButtons(QDialogButtonBox.StandardButton.Cancel | QDialogButtonBox.StandardButton.Ok)
        buttons.accepted.connect(self._on_ok_pressed)
        buttons.rejected.connect(self._on_cancel_pressed)
        self._ok_button = buttons.button(QDialogButtonBox.StandardButton.Ok)

        if multiline:
            editor = QCommentTextBox(
                parent=self, textconfirmed_callback=self._on_ok_pressed, textchanged_callback=self._evaluate
            )
            editor.setFont(Conf.disasm_font)
            self.text = editor.toPlainText
            self.setText = editor.setPlainText

            editor.setPlainText(base_text)
            editor.selectAll()

        else:
            editor = QLineEdit(self)
            editor.setFont(Conf.disasm_font)
            self.text = editor.text
            self.setText = editor.setText
            editor.returnPressed.connect(self._on_ok_pressed)
            editor.textChanged.connect(self._evaluate)

            editor.setText(base_text)
            editor.setFocus()
            editor.selectAll()

        layout = QVBoxLayout()
        layout.addWidget(editor)
        layout.addWidget(buttons)

        self.setLayout(layout)

    def _on_ok_pressed(self) -> None:
        self.close()

    def _on_cancel_pressed(self) -> None:
        self.main_result = []
        self.side_result = []
        self.close()

    def _evaluate(self) -> None:
        text = self.text()

        result = None
        side_result = None
        try:
            side_typedefs = {}
            defs, typedefs = sim_type.parse_file(
                text, predefined_types=self._predefined_types, arch=self.arch, side_effect_types=side_typedefs
            )
            result = list(typedefs.items()) + list(defs.items())
            side_result = [
                (name, ty) for name, ty in side_typedefs.items() if name not in typedefs and name not in defs
            ]
        except PycParseErrorCls:
            pass

        # hack. idk why our pycparser config will accept `typedef int` as the same as `int`
        if result is None and not text.strip().startswith("typedef"):
            try:
                side_typedefs = {}
                ty, name = sim_type.parse_type_with_name(
                    text, predefined_types=self._predefined_types, arch=self.arch, side_effect_types=side_typedefs
                )
                result = [(name, ty)]
                side_result = [(name2, ty) for name2, ty in side_typedefs.items() if name2 != name]
            except PycParseErrorCls:
                pass

        if not result or (self._editing_single and len(result) != 1):
            self._ok_button.setEnabled(False)
            self.main_result = []
            self.side_result = []
        else:
            self._ok_button.setEnabled(True)
            self.main_result = result
            assert side_result is not None
            self.side_result = side_result


def edit_field(ty, field, predefined_types=None) -> bool:
    if isinstance(ty, sim_type.SimStruct):
        fields = ty.fields
    elif isinstance(ty, sim_type.SimUnion):
        fields = ty.members
    else:
        raise TypeError(f"Can't edit a field of a {type(ty)}")

    if type(fields) is not OrderedDict:
        raise TypeError(f"Struct or union's fields are of type {type(fields)} - that's bad")
    fields_list = list(fields.items())

    if isinstance(field, int):
        if not 0 <= field < len(fields):
            raise IndexError(field)
        fieldno = field
    elif isinstance(field, str):
        try:
            fieldno = [i for i, (name, _) in fields_list if name == field][0]
        except IndexError:
            raise KeyError(field) from None
    else:
        raise TypeError(f"Field specifier is a {type(field)} - that's bad")

    name, subty = fields_list[fieldno]
    text = subty.c_repr(name=name)
    dialog = CTypeEditor(None, ty._arch, text, multiline=False, editing_single=name, predefined_types=predefined_types)
    dialog.exec_()
    if not dialog.result:
        return False
    name2, subty = dialog.result[0]
    if name2 is not None:
        if name != name2 and name2 in fields:
            QMessageBox.warning(None, "Duplicate field name", f"The name {name2} is already used")
        else:
            name = name2
    fields_list[fieldno] = (name, subty)

    if isinstance(ty, sim_type.SimStruct):
        ty.fields = OrderedDict(fields_list)
    elif isinstance(ty, sim_type.SimUnion):
        ty.members = OrderedDict(fields_list)
    else:
        raise TypeError(f"Can't edit a field of a {type(ty)}")

    return True



from __future__ import annotations

from typing import TYPE_CHECKING

from angrmanagement.ui.icons import icon

from .toolbar import Toolbar, ToolbarAction

if TYPE_CHECKING:
    from angrmanagement.ui.main_window import MainWindow


class FileToolbar(Toolbar):
    def __init__(self, main_window: MainWindow) -> None:
        super().__init__(main_window, "File")

        self.actions = [
            ToolbarAction(
                icon("file-open"),
                "Open File",
                "Open a new file for analysis",
                main_window.open_file_button,
            ),
            ToolbarAction(
                icon("file-save"),
                "Save",
                "Save angr database",
                main_window.save_database,
            ),
        ]



from __future__ import annotations

from typing import TYPE_CHECKING

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QProgressDialog

if TYPE_CHECKING:
    from angrmanagement.ui.main_window import MainWindow


class ProgressDialog(QProgressDialog):
    """ProgressDialog is a dialog that shows the progress of a job on top of the main window."""

    main_window: MainWindow

    def __init__(self, main_window: MainWindow):
        super().__init__("Waiting...", "Cancel", 0, 100, parent=main_window)
        self.setWindowFlags(self.windowFlags() & ~Qt.WindowType.WindowContextHelpButtonHint)
        self.setModal(True)
        self.setMinimumDuration(0)
        self.canceled.disconnect()
        self.reset()



from __future__ import annotations

import json
from typing import TYPE_CHECKING

from PySide6.QtCore import Qt
from PySide6.QtGui import QColor
from PySide6.QtWidgets import QApplication, QFileDialog, QInputDialog, QLineEdit, QMessageBox

from angrmanagement.errors import InvalidURLError, UnexpectedStatusCodeError
from angrmanagement.logic import GlobalInfo
from angrmanagement.logic.threads import gui_thread_schedule_async
from angrmanagement.plugins.base_plugin import BasePlugin
from angrmanagement.utils.io import download_url, isurl

from .afl_qemu_bitmap import AFLQemuBitmap
from .multi_trace import MultiTrace
from .qtrace_viewer import QTraceViewer
from .trace_statistics import TraceStatistics

if TYPE_CHECKING:
    from angrmanagement.data.object_container import ObjectContainer
    from angrmanagement.ui.workspace import Workspace


class TraceViewer(BasePlugin):
    def __init__(self, workspace: Workspace) -> None:
        super().__init__(workspace)

        self.workspace.main_instance.register_container(
            "trace", lambda: None, TraceStatistics | None, "The current trace"
        )
        self.workspace.main_instance.register_container(
            "multi_trace", lambda: None, MultiTrace | None, "The current set of multiple traces"
        )

        # Register event callbacks
        # self.trace.am_subscribe(self._on_trace_updated)
        self.multi_trace.am_subscribe(self._on_trace_updated)

        self._viewers = []

    def teardown(self) -> None:
        # I don't really know a better way to do this. tbh allowing arbitrary widget additions is probably intractable
        for trace_viewer in self._viewers:
            trace_viewer.hide()

    #
    # Forwarding properties
    #

    @property
    def trace(self) -> ObjectContainer | TraceStatistics | None:
        return self.workspace.main_instance.trace

    @property
    def multi_trace(self) -> ObjectContainer | MultiTrace | None | AFLQemuBitmap:
        return self.workspace.main_instance.multi_trace

    #
    # Event handlers
    #

    def _on_trace_updated(self) -> None:
        # redraw disassembly view
        view = self.workspace.view_manager.first_view_in_category("disassembly")
        if view is not None:
            view.redraw_current_graph()
        # refresh function table
        view = self.workspace.view_manager.first_view_in_category("functions")
        if view is not None:
            view.refresh()

    URL_ACTIONS = ["openbitmap", "opentrace"]

    def handle_url_action(self, action, kwargs):
        if action == "openbitmap":
            try:
                base = int(kwargs["base"], 16)
            except (ValueError, KeyError):
                base = None
            gui_thread_schedule_async(GlobalInfo.main_window.bring_to_front)
            gui_thread_schedule_async(
                self.open_bitmap_multi_trace,
                args=(kwargs["path"],),
                kwargs={
                    "base_addr": base,
                },
            )
        elif action == "opentrace":
            try:
                base = int(kwargs["base"], 16)
            except (ValueError, KeyError):
                base = None
            path = kwargs["path"]
            gui_thread_schedule_async(GlobalInfo.main_window.bring_to_front)
            gui_thread_schedule_async(
                self.add_trace,
                kwargs={
                    "trace_path": path,
                    "base_addr": base,
                },
            )
        else:
            raise ValueError("Trace plugin cannot handle url action " + action)

    #
    # features for the disassembly view
    #

    GRAPH_TRACE_LEGEND_WIDTH = 30
    GRAPH_TRACE_LEGEND_SPACING = 20

    def instrument_disassembly_view(self, dview) -> None:
        trace_viewer = QTraceViewer(self.workspace, dview, parent=dview)
        self._viewers.append(trace_viewer)

        dview.layout().addWidget(trace_viewer)
        trace_viewer.hide()

    def color_block(self, addr: int):
        if not self.multi_trace.am_none:
            if isinstance(self.multi_trace.am_obj, MultiTrace) and not self.multi_trace.is_active_tab:
                return None
            return self.multi_trace.get_hit_miss_color(addr)
        return None

    def handle_click_block(self, qblock, event) -> bool:
        btn = event.button()

        if (
            QApplication.keyboardModifiers() == Qt.KeyboardModifier.ControlModifier
            and btn == Qt.MouseButton.RightButton
            and self.multi_trace is not None
            and self.multi_trace.am_obj is not None
        ):
            the_trace = self.multi_trace.get_any_trace(qblock.addr)
            if the_trace is not None:
                self.trace.am_obj = TraceStatistics(self.workspace, the_trace, self.multi_trace.base_addr)
                self.trace.am_event()
                return True

        return False

    def draw_insn(self, qinsn, painter) -> None:
        # legend
        if (
            not self.multi_trace.am_none
            and isinstance(self.multi_trace.am_obj, MultiTrace)
            and not self.multi_trace.is_active_tab
        ):
            return  # skip
        strata = self._gen_strata(qinsn.insn.addr)
        if strata is not None:
            legend_x = 0 - self.GRAPH_TRACE_LEGEND_WIDTH - self.GRAPH_TRACE_LEGEND_SPACING
            for i, w in strata:
                color = self.trace.get_mark_color(qinsn.insn.addr, i)
                painter.setPen(color)
                painter.setBrush(color)
                painter.drawRect(legend_x, 0, w, qinsn.height)
                legend_x += w

    def _gen_strata(self, addr: int):
        if not self.trace.am_none:
            # count is cached in trace.
            count = self.trace.get_count(addr)

            if count > self.GRAPH_TRACE_LEGEND_WIDTH:
                jump = count / self.GRAPH_TRACE_LEGEND_WIDTH
                return [(int(jump * i), 1) for i in range(self.GRAPH_TRACE_LEGEND_WIDTH)]
            elif count > 0:
                width = self.GRAPH_TRACE_LEGEND_WIDTH // count
                remainder = self.GRAPH_TRACE_LEGEND_WIDTH % count
                return [(i, width + 1) for i in range(remainder)] + [(i, width) for i in range(remainder, count)]

        return None

    #
    # features for the functions view
    #

    def color_func(self, func):
        if not self.multi_trace.am_none:
            if isinstance(self.multi_trace.am_obj, MultiTrace):
                if self.multi_trace.is_active_tab:
                    return self.multi_trace.get_percent_color(func)
            elif isinstance(self.multi_trace.am_obj, AFLQemuBitmap):
                return self.multi_trace.get_percent_color(func)
            else:
                # you should not reach here
                raise RuntimeError("Impossible happened")

        if not self.trace.am_none:
            if func.addr in self.trace.func_addr_in_trace:
                return QColor(0xF0, 0xE7, 0xDA)
            return QColor(0xEE, 0xEE, 0xEE)
        return None

    FUNC_COLUMNS = ("Coverage",)

    def extract_func_column(self, func, idx: int):
        assert idx == 0
        if self.multi_trace.am_none:
            cov = 0
            rend = ""
        else:
            cov = self.multi_trace.get_coverage(func)
            rend = f"{cov:.2f}%"

        return cov, rend

    #
    # features for loading traces!
    #

    MENU_BUTTONS = [
        "Open/Add trace...",
        "Clear trace",
        "Open AFL bitmap...",
        "Open inverted AFL bitmap...",
        "Reset AFL bitmap",
    ]
    ADD_TRACE_ID = 0
    RESET_TRACE_ID = 1
    OPEN_AFL_BITMAP_ID = 2
    OPEN_AFL_BITMAP_INVERTED_ID = 3
    RESET_AFL_BITMAP = 4

    def handle_click_menu(self, idx: int) -> None:
        if idx < 0 or idx >= len(self.MENU_BUTTONS):
            return

        if self.workspace.main_instance.project.am_none:
            return

        mapping = {
            self.ADD_TRACE_ID: self.add_trace,
            self.RESET_TRACE_ID: self.reset_trace,
            self.OPEN_AFL_BITMAP_ID: self.open_bitmap_multi_trace,
            self.OPEN_AFL_BITMAP_INVERTED_ID: self.open_inverted_bitmap_multi_trace,
            self.RESET_AFL_BITMAP: self.reset_bitmap,
        }

        mapping.get(idx)()

    def add_trace(self, trace_path=None, base_addr=None) -> None:
        trace, base_addr = self._open_trace(trace_path, base_addr)
        if trace is None or base_addr is None:
            return

        self._add_trace(trace, base_addr)

    def reset_trace(self) -> None:
        self.trace.am_obj = None
        self.trace.am_event()
        self.multi_trace.am_obj = None
        self.multi_trace.am_event()

    def open_bitmap_multi_trace(self, trace_path=None, base_addr=None) -> None:
        r = self._open_bitmap_multi_trace(trace_path, base_addr)
        if r is None:
            return
        trace, base_addr = r
        self.multi_trace.am_obj = AFLQemuBitmap(self.workspace, trace, base_addr)
        self.multi_trace.am_event()

    def open_inverted_bitmap_multi_trace(self, trace_path=None, base_addr=None) -> None:
        r = self._open_bitmap_multi_trace(trace_path, base_addr)
        if r is None:
            return
        trace, base_addr = r
        self.multi_trace.am_obj = AFLQemuBitmap(self.workspace, trace, base_addr, bits_inverted=True)
        self.multi_trace.am_event()

    def reset_bitmap(self) -> None:
        self.multi_trace.am_obj = None
        self.multi_trace.am_event()

    def _open_bitmap_multi_trace(self, trace_path, base_addr):
        if trace_path is None:
            trace_path = self._open_trace_dialog(tfilter="")
            if trace_path is None:
                return None

        if base_addr is None:
            base_addr = self._open_baseaddr_dialog(0x4000000000)
            if base_addr is None:
                return None

        if isurl(trace_path):
            try:
                trace = download_url(trace_path, parent=self.workspace._main_window, to_file=False)
            except InvalidURLError:
                QMessageBox.critical(
                    self.workspace._main_window,
                    "Downloading failed",
                    "angr management failed to download the file. The URL is invalid.",
                )
                return None
            except UnexpectedStatusCodeError as ex:
                QMessageBox.critical(
                    self.workspace._main_window,
                    "Downloading failed",
                    "angr management failed to retrieve the header of the file. "
                    f"The HTTP request returned an unexpected status code {ex.status_code}.",
                )
                return None
        else:
            with open(trace_path, "rb") as f:
                trace = f.read()

        return trace, base_addr

    def _open_trace(self, trace_path, base_addr):
        if trace_path is None:
            return self._open_json_trace_dialog()
        else:
            if isurl(trace_path):
                try:
                    trace_bytes = download_url(trace_path, parent=self.workspace._main_window, to_file=False)
                    trace = json.loads(trace_bytes)
                except InvalidURLError:
                    QMessageBox.critical(
                        self.workspace._main_window,
                        "Downloading failed",
                        "angr management failed to download the file. The URL is invalid.",
                    )
                    trace = None
                except UnexpectedStatusCodeError as ex:
                    QMessageBox.critical(
                        self.workspace._main_window,
                        "Downloading failed",
                        "angr management failed to retrieve the header of the file. "
                        f"The HTTP request returned an unexpected status code {ex.status_code}.",
                    )
                    trace = None
            else:
                with open(trace_path) as f:
                    trace = json.load(f)

            if base_addr is None:
                base_addr = self._open_baseaddr_dialog(0x4000000000)

            return trace, base_addr

    def _open_trace_dialog(self, tfilter):
        file_path, _ = QFileDialog.getOpenFileName(None, "Open a trace", "", tfilter)
        try:
            with open(file_path, "rb") as f:
                f.read(1)
        except FileNotFoundError:
            return None

        return file_path

    def _open_baseaddr_dialog(self, default_base):
        base_addr, _ = QInputDialog.getText(
            None, "Input Trace Base Address", "Base Address:", QLineEdit.Normal, hex(default_base)
        )
        try:
            return int(base_addr, 16)
        except ValueError:
            return None

    def _open_json_trace_dialog(self) -> tuple[list[int] | None, int | None]:
        # project = self.workspace.instance.project
        trace_file_name = self._open_trace_dialog(tfilter="json (*.json)")

        if trace_file_name is None:
            return None, None

        with open(trace_file_name) as f:
            trace = json.load(f)

        if not isinstance(trace, dict):
            QMessageBox.critical(
                self.workspace._main_window,
                "Incorrect trace format",
                "Failed to open the JSON trace. We expect the JSON trace to be a dict.",
            )
            return None, None
        elif "bb_addrs" not in trace:
            QMessageBox.critical(
                self.workspace._main_window,
                "Incorrect trace format",
                "Failed to open the JSON trace. We expect the JSON trace to contain the field 'bb_addrs'.",
            )
            return None, None
        elif not isinstance(trace["bb_addrs"], list):
            QMessageBox.critical(
                self.workspace._main_window,
                "Incorrect trace format",
                "Failed to open the JSON trace. We expect the JSON trace bb_addrs field to be a list of integers.",
            )
            return None, None

        base_addr = self._open_baseaddr_dialog(0x4000000000)

        return trace, base_addr

    def _add_trace(self, trace, base_addr) -> None:
        if self.multi_trace.am_obj is None:
            self.multi_trace.am_obj = MultiTrace(self.workspace)
        self.trace.am_obj = self.multi_trace.am_obj.add_trace(trace, base_addr)
        self.multi_trace.am_event()
        self.trace.am_event()



# pylint:disable=missing-class-docstring,wrong-import-order
from __future__ import annotations

import os
import sys
import unittest

import angr
from common import AngrManagementTestCase, test_location

from angrmanagement.ui.views import DisassemblyView


class TestQDisassemblyGraph(AngrManagementTestCase):
    def test_disassembly_level_change(self):
        main = self.main
        binpath = os.path.join(test_location, "x86_64", "fauxware")
        main.workspace.main_instance.project.am_obj = angr.Project(binpath, auto_load_libs=False)
        main.workspace.main_instance.project.am_event()
        main.workspace.job_manager.join_all_jobs()

        func = main.workspace.main_instance.project.kb.functions["main"]
        assert func is not None

        # load the disassembly view in graph mode
        disasm_view = main.workspace._get_or_create_view("disassembly", DisassemblyView)
        disasm_view.display_disasm_graph()
        disasm_view.display_function(func)
        main.workspace.job_manager.join_all_jobs()

        # find the entry block in the graph view
        flow_graph = disasm_view.flow_graph
        entry_block = next((b for b in flow_graph.blocks if b.addr == func.addr), None)
        assert entry_block is not None
        assert entry_block.disasm._include_ir is False

        # change disassembly level to LifterIR
        disasm_view.set_disassembly_level_lifter_ir()
        main.workspace.job_manager.join_all_jobs()

        # blocks list is rebuilt by reload() — re-fetch the entry block
        entry_block = next((b for b in flow_graph.blocks if b.addr == func.addr), None)
        assert entry_block is not None
        assert entry_block.disasm._include_ir is True


if __name__ == "__main__":
    unittest.main(argv=sys.argv)



from __future__ import annotations

from typing import TYPE_CHECKING

from PySide6.QtCore import QSize, Qt
from PySide6.QtGui import QPainter, QPen
from PySide6.QtWidgets import QFrame, QHBoxLayout, QLabel, QLineEdit, QScrollArea, QVBoxLayout, QWidget

from angrmanagement.config import Conf

from .qast_viewer import QASTViewer

if TYPE_CHECKING:
    from angrmanagement.ui.workspace import Workspace


class AddressPiece:
    __slots__ = ["address"]

    def __init__(self, address) -> None:
        self.address = address


class NewLinePiece:
    pass


class QMemoryView(QWidget):
    def __init__(self, state, workspace: Workspace, parent=None) -> None:
        super().__init__(parent)
        self.workspace = workspace

        self.state = state
        self.cols = None
        self.rows = None

        # The current address being displayed. Must be set through .address
        self._address = None

        self._objects = []

    @property
    def address(self):
        return self._address

    @address.setter
    def address(self, v) -> None:
        if v != self._address:
            self._address = v
            self._reload_objects()

    def paintEvent(self, event):
        if self.address is None:
            return

        MARGIN_LEFT = 5
        MARGIN_TOP = 5
        LINE_MARGIN = 3

        painter = QPainter(self)

        painter.setPen(QPen(Qt.black, 1))
        painter.setFont(Conf.symexec_font)

        x = MARGIN_LEFT
        y = MARGIN_TOP

        for obj in self._objects:
            obj_type = type(obj)

            if obj_type is NewLinePiece:
                # carriage return
                x = MARGIN_LEFT
                y += Conf.symexec_font_height + LINE_MARGIN
            elif obj_type is AddressPiece:
                # address
                addr_str = f"{obj.address:08x}"
                painter.drawText(x, y + Conf.symexec_font_ascent, addr_str)
                x += Conf.symexec_font_width * len(addr_str)
                x += 7
            elif obj_type is QASTViewer:
                # AST viewer
                obj.x = x
                obj.y = y
                obj.paint(painter)

                x += obj.width + 2
            else:
                raise TypeError(f"paintEvent(): Unsupported object type {obj_type}.")

    def _reload_objects(self) -> None:
        """
        Reload addresses and text pieces to be displayed.

        :return: None
        """

        objects = []

        addr_base = self.address
        for row in range(self.rows):
            addr = addr_base + row * self.cols

            # address
            addr_piece = AddressPiece(addr)
            objects.append(addr_piece)

            # QASTViewer objects
            for col in range(self.cols):
                data = self.state.memory.load(addr + col, 1, inspect=False, disable_actions=True)
                ast_viewer = QASTViewer(data, workspace=self.workspace, custom_painting=True, display_size=False)
                objects.append(ast_viewer)

            # end of the line
            newline_piece = NewLinePiece()
            objects.append(newline_piece)

        self._objects = objects


class QMemoryViewer(QFrame):
    def __init__(self, state, parent, workspace: Workspace) -> None:
        super().__init__(parent)
        self.workspace = workspace

        self._txt_addr: QLineEdit
        self._view: QMemoryView

        self._addr = None  # the address to display
        self.state = state

        self._init_widgets()

        self.state.am_subscribe(self._watch_state)

    @property
    def addr(self):
        return self._addr

    @addr.setter
    def addr(self, v) -> None:
        if self._addr != v:
            self._addr = v

            self.reload()

    #
    # Overridden methods
    #

    def sizeHint(self):
        return QSize(100, 100)

    #
    # Public methods
    #

    def reload(self) -> None:
        if self.state.am_none:
            return

        if self.addr is None:
            return

        self._refresh_memory_view()

    #
    # Event handlers
    #

    def _on_address_entered(self) -> None:
        address_str = self._txt_addr.text()

        try:
            address = int(address_str, 16)
        except ValueError:
            return

        self.addr = address

    #
    # Private methods
    #

    def _init_widgets(self) -> None:
        layout = QVBoxLayout()

        # address

        lbl_addr = QLabel()
        lbl_addr.setText("Address")

        txt_addr = QLineEdit()
        txt_addr.returnPressed.connect(self._on_address_entered)
        self._txt_addr = txt_addr

        top_layout = QHBoxLayout()
        top_layout.addWidget(lbl_addr)
        top_layout.addWidget(txt_addr)

        self._view = QMemoryView(self.state, self.workspace)

        area = QScrollArea()
        area.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)
        area.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)
        area.setWidgetResizable(True)

        area.setWidget(self._view)

        layout.addLayout(top_layout)
        layout.addWidget(area)
        layout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(layout)

    def _refresh_memory_view(self) -> None:
        self._view.cols = 16
        self._view.rows = 10
        self._view.address = self.addr

        self._view.repaint()

    def _watch_state(self, **kwargs) -> None:
        self.reload()



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from .job import InstanceJob

if TYPE_CHECKING:
    from angr.knowledge_plugins.functions.function import Function

    from angrmanagement.data.instance import Instance
    from angrmanagement.logic.jobmanager import JobContext

log = logging.getLogger(__name__)


class LLMPreloadCalleesJob(InstanceJob):
    """
    Job that auto-decompiles and LLM-refines callees of a given function.
    This pre-loads refined decompilations for functions called by the target,
    so that when the user navigates into them the results are already available.
    """

    def __init__(self, instance: Instance, function: Function, on_finish=None) -> None:
        super().__init__("LLM Preload Callees", instance, on_finish=on_finish)
        self.function = function

    def run(self, ctx: JobContext) -> None:
        project = self.instance.project.am_obj
        if project is None:
            return

        llm_client = project.llm_client
        if llm_client is None:
            return

        kb = self.instance.kb
        callgraph = kb.functions.callgraph

        if self.function.addr not in callgraph:
            return

        # get callees
        callee_addrs = list(callgraph.successors(self.function.addr))

        if not callee_addrs:
            return

        # filter out simprocedures and PLT stubs
        callees = []
        for addr in callee_addrs:
            if addr not in kb.functions:
                continue
            func = kb.functions[addr]
            if func.is_simprocedure or func.is_plt:
                continue
            callees.append(func)

        if not callees:
            return

        total = len(callees)
        for i, callee in enumerate(callees):
            if self.cancelled:
                return

            pct = int((i / total) * 100)
            ctx.set_progress(pct, f"Refining callee {callee.name} ({i + 1}/{total})")

            # decompile if not already cached
            dec_cache = kb.decompilations.get((callee.addr, "pseudocode"))
            if dec_cache is None or dec_cache.codegen is None:
                try:
                    decompiler = project.analyses.Decompiler(
                        callee,
                        flavor="pseudocode",
                        variable_kb=self.instance.pseudocode_variable_kb,
                    )
                    kb.decompilations[(callee.addr, "pseudocode")] = decompiler.cache
                    dec_cache = decompiler.cache
                except Exception:  # pylint: disable=broad-except
                    log.debug("Failed to decompile callee %s", callee.name, exc_info=True)
                    continue

            if dec_cache is None or dec_cache.codegen is None:
                continue

            # run LLM refinement
            dec = project.analyses.Decompiler(
                callee,
                variable_kb=self.instance.pseudocode_variable_kb,
                decompile=False,
            )
            dec.codegen = dec_cache.codegen
            changed = dec.llm_refine()
            if changed:
                dec_cache.codegen = dec.codegen

        ctx.set_progress(100, "Done")



"""
Test cases for DisassemblyView.
"""

from __future__ import annotations

import contextlib
import unittest
from unittest.mock import MagicMock, call, patch

from common import ProjectOpenTestCase
from PySide6.QtCore import QPointF, Qt
from PySide6.QtGui import QKeyEvent
from PySide6.QtWidgets import QDialog

from angrmanagement.ui.dialogs.assemble_patch import AssemblePatchDialog
from angrmanagement.ui.dialogs.hook import HookDialog
from angrmanagement.ui.dialogs.jumpto import JumpTo
from angrmanagement.ui.dialogs.new_state import NewState
from angrmanagement.ui.dialogs.rename_label import RenameLabel
from angrmanagement.ui.dialogs.set_comment import SetComment
from angrmanagement.ui.dialogs.xref import XRefDialog
from angrmanagement.ui.views.disassembly_view import DisassemblyView


class TestDisassemblyViewBase(ProjectOpenTestCase):
    """Base class with shared view setup for DisassemblyView tests."""

    def setUp(self):
        super().setUp()

        self.disasm_view = DisassemblyView(self.workspace, "center", self.instance)

        # Mock disasm property to avoid None access errors in dialog constructors
        mock_disasm = MagicMock()
        mock_disasm.kb = self.instance.project.kb
        self.disasm_view._flow_graph.disasm = mock_disasm


class TestPopupDialogs(TestDisassemblyViewBase):
    """Test popup_*_dialog methods that create and show dialogs."""

    def setUp(self):
        super().setUp()

        self._show_called = False
        self._exec_called = False
        self._dialog_instance = None

        def mock_show(dialog_self):
            self._show_called = True
            self._dialog_instance = dialog_self

        def mock_exec(dialog_self, *_args, **_kwargs):
            self._exec_called = True
            self._dialog_instance = dialog_self
            return QDialog.DialogCode.Accepted

        self._show_patcher = patch("PySide6.QtWidgets.QDialog.show", mock_show)
        self._exec_patcher = patch("PySide6.QtWidgets.QDialog.exec_", mock_exec)

        self._show_patcher.start()
        self._exec_patcher.start()

        self.addCleanup(self._show_patcher.stop)
        self.addCleanup(self._exec_patcher.stop)

    def test_popup_newstate_dialog(self):
        """Test that DisassemblyView.popup_newstate_dialog() opens NewState dialog."""

        with patch.object(self.disasm_view, "_instruction_address_in_selection", return_value=0x1000):
            self.disasm_view.popup_newstate_dialog()

            assert self._exec_called, "NewState dialog should use .exec_() (modal)"
            assert isinstance(self._dialog_instance, NewState)
            self._dialog_instance.close()

    def test_popup_patch_dialog(self):
        """Test that DisassemblyView.popup_patch_dialog() opens AssemblePatchDialog (modal)."""

        mock_dialog_instance = MagicMock(spec=AssemblePatchDialog)

        with patch("angrmanagement.ui.views.disassembly_view.AssemblePatchDialog", return_value=mock_dialog_instance):
            self.disasm_view._insn_addr_on_context_menu = 0x1418
            self.disasm_view.popup_patch_dialog()
            mock_dialog_instance.exec_.assert_called_once()

    def test_popup_hook_dialog(self):
        """Test that DisassemblyView.popup_hook_dialog() opens HookDialog (modal)."""

        with patch.object(self.disasm_view, "_instruction_address_in_selection", return_value=0x1000):
            self.disasm_view.popup_hook_dialog()

            assert self._exec_called, "HookDialog should use .exec_() (modal)"
            assert isinstance(self._dialog_instance, HookDialog)
            self._dialog_instance.close()

    def test_popup_xref_dialog_with_variable(self):
        """Test that popup_xref_dialog creates and shows XRefDialog with variable."""
        mock_variable = MagicMock()
        self.disasm_view.variable_manager = MagicMock()

        self.disasm_view.popup_xref_dialog(addr=0x1000, variable=mock_variable)

        assert self._exec_called, "popup_xref_dialog should open XRefDialog with .exec_()"
        assert isinstance(self._dialog_instance, XRefDialog)
        self._dialog_instance.close()

    def test_popup_xref_dialog_with_dst_addr(self):
        """Test that popup_xref_dialog creates and shows XRefDialog with destination address."""
        self.disasm_view.popup_xref_dialog(addr=0x1000, dst_addr=0x2000)

        assert self._exec_called, "popup_xref_dialog should open XRefDialog with .exec_()"
        assert isinstance(self._dialog_instance, XRefDialog)
        self._dialog_instance.close()

    def test_parse_operand_and_popup_xref_dialog_with_constant(self):
        """Test that parse_operand_and_popup_xref_dialog handles constant operand."""
        mock_operand = MagicMock()
        mock_operand.is_constant = True
        mock_operand.constant_value = 0x2000
        mock_operand.variable = None

        self.disasm_view.parse_operand_and_popup_xref_dialog(0x1000, mock_operand)

        assert self._exec_called, "parse_operand_and_popup_xref_dialog should open XRefDialog with .exec_()"
        assert isinstance(self._dialog_instance, XRefDialog)
        self._dialog_instance.close()

    def test_parse_operand_and_popup_xref_dialog_with_variable(self):
        """Test that parse_operand_and_popup_xref_dialog handles variable operand."""
        mock_variable = MagicMock()
        mock_operand = MagicMock()
        mock_operand.variable = mock_variable
        self.disasm_view.variable_manager = MagicMock()

        self.disasm_view.parse_operand_and_popup_xref_dialog(0x1000, mock_operand)

        assert self._exec_called, "parse_operand_and_popup_xref_dialog should open XRefDialog with .exec_()"
        assert isinstance(self._dialog_instance, XRefDialog)
        self._dialog_instance.close()

    def test_parse_operand_and_popup_xref_dialog_with_constant_memory(self):
        """Test that parse_operand_and_popup_xref_dialog handles constant_memory operand."""
        mock_operand = MagicMock()
        mock_operand.variable = None
        mock_operand.is_constant = False
        mock_operand.is_constant_memory = True
        mock_operand.constant_memory_value = 0x3000

        self.disasm_view.parse_operand_and_popup_xref_dialog(0x1000, mock_operand)

        assert self._exec_called, "parse_operand_and_popup_xref_dialog should open XRefDialog with .exec_()"
        assert isinstance(self._dialog_instance, XRefDialog)
        self._dialog_instance.close()

    def test_parse_operand_and_popup_xref_dialog_with_none_operand(self):
        """Test that parse_operand_and_popup_xref_dialog handles None operand gracefully."""
        self.disasm_view.parse_operand_and_popup_xref_dialog(0x1000, None)

        assert not self._exec_called, "No dialog should be created when operand is None"
        assert self._dialog_instance is None

    def test_popup_jumpto_dialog(self):
        """Test that popup_jumpto_dialog creates and shows JumpTo dialog (non-modal)."""
        self.disasm_view.popup_jumpto_dialog()

        assert self._show_called, "popup_jumpto_dialog should open JumpTo with .show()"
        assert isinstance(self._dialog_instance, JumpTo)
        self._dialog_instance.close()

    def test_popup_comment_dialog(self):
        """Test that popup_comment_dialog creates and shows SetComment dialog (modal)."""
        with patch.object(self.disasm_view, "_instruction_address_in_selection", return_value=0x1000):
            self.disasm_view.popup_comment_dialog()

            assert self._exec_called, "popup_comment_dialog should open SetComment with .exec_()"
            assert isinstance(self._dialog_instance, SetComment)
            self._dialog_instance.close()

    def test_popup_rename_label_dialog(self):
        """Test that popup_rename_label_dialog creates and shows RenameLabel dialog (modal)."""
        with patch.object(self.disasm_view, "_address_in_selection", return_value=(0x1000, False)):
            self.disasm_view.popup_rename_label_dialog()

            assert self._exec_called, "popup_rename_label_dialog should open RenameLabel with .exec_()"
            assert isinstance(self._dialog_instance, RenameLabel)
            self._dialog_instance.close()


class TestKeyboardShortcuts(TestDisassemblyViewBase):
    """Test keyboard shortcut routing in keyPressEvent."""

    def test_g_key_calls_popup_jumpto(self):
        """Test that pressing 'g' key calls popup_jumpto_dialog."""
        with patch.object(self.disasm_view, "popup_jumpto_dialog") as mock_popup:
            key_event = QKeyEvent(QKeyEvent.Type.KeyPress, Qt.Key.Key_G, Qt.KeyboardModifier.NoModifier)
            self.disasm_view.keyPressEvent(key_event)
            mock_popup.assert_called_once()

    def test_semicolon_key_calls_popup_comment(self):
        """Test that pressing ';' key calls popup_comment_dialog."""
        with patch.object(self.disasm_view, "popup_comment_dialog") as mock_popup:
            key_event = QKeyEvent(QKeyEvent.Type.KeyPress, Qt.Key.Key_Semicolon, Qt.KeyboardModifier.NoModifier)
            self.disasm_view.keyPressEvent(key_event)
            mock_popup.assert_called_once()

    def test_n_key_calls_popup_rename_label(self):
        """Test that pressing 'n' key in control calls popup_rename_label_dialog."""
        control = self.disasm_view._flow_graph

        with patch.object(self.disasm_view, "popup_rename_label_dialog") as mock_popup:
            key_event = QKeyEvent(QKeyEvent.Type.KeyPress, Qt.Key.Key_N, Qt.KeyboardModifier.NoModifier)
            control.keyPressEvent(key_event)
            mock_popup.assert_called_once()

    @staticmethod
    @contextlib.contextmanager
    def _mock_infodock_selection(control, operands=None, labels=None, variables=None, block_tree_node=None):
        """
        Context manager to mock infodock selections.
        """
        operands = operands if operands is not None else set()
        labels = labels if labels is not None else set()
        variables = variables if variables is not None else set()

        if block_tree_node is None:
            block_tree_node = MagicMock()
            block_tree_node.am_none = True

        with (
            patch.object(control.infodock, "selected_operands", operands),
            patch.object(control.infodock, "selected_labels", labels),
            patch.object(control.infodock, "selected_variables", variables),
            patch.object(control.infodock, "selected_block_tree_node", block_tree_node),
        ):
            yield

    def _test_x_key_calls_method(self, method_name, expected_call, **selection_kwargs):
        """
        Test that X key calls the specified method with expected arguments.
        """
        control = self.disasm_view._flow_graph

        with (
            self._mock_infodock_selection(control, **selection_kwargs),
            patch.object(self.disasm_view, method_name) as mock_method,
        ):
            control.keyPressEvent(QKeyEvent(QKeyEvent.Type.KeyPress, Qt.Key.Key_X, Qt.KeyboardModifier.NoModifier))
            assert mock_method.call_args == expected_call

    def test_x_key_with_operand_calls_parse_operand_xref(self):
        """Test that pressing 'x' key with operand selected calls parse_operand_and_popup_xref_dialog."""
        control = self.disasm_view._flow_graph

        mock_operand = MagicMock()
        mock_operand.is_constant = True
        mock_operand.constant_value = 0x2000
        mock_operand.variable = None

        mock_block = MagicMock()
        mock_insn = MagicMock()
        mock_insn.get_operand.return_value = mock_operand
        mock_block.addr_to_insns = {0x1000: mock_insn}
        control._insaddr_to_block[0x1000] = mock_block

        self._test_x_key_calls_method(
            "parse_operand_and_popup_xref_dialog",
            call(0x1000, mock_operand),
            operands={(0x1000, 0)},
        )

    def test_x_key_with_label_calls_popup_xref(self):
        """Test that pressing 'x' key with label selected calls popup_xref_dialog."""
        self._test_x_key_calls_method("popup_xref_dialog", call(addr=0x3000, dst_addr=0x3000), labels={0x3000})

    def test_x_key_with_variable_calls_popup_xref(self):
        """Test that pressing 'x' key with variable selected calls popup_xref_dialog."""
        mock_variable = MagicMock()
        self._test_x_key_calls_method(
            "popup_xref_dialog", call(addr=0, variable=mock_variable), variables={mock_variable}
        )

    def test_x_key_with_function_header_calls_popup_xref(self):
        """Test that pressing 'x' key with function header selected calls popup_xref_dialog."""
        mock_block_tree_node = MagicMock()
        mock_block_tree_node.am_none = False
        mock_block_tree_node.am_obj = ("func_name", 0x4000)

        self._test_x_key_calls_method(
            "popup_xref_dialog", call(addr=0x4000, dst_addr=0x4000), block_tree_node=mock_block_tree_node
        )

    def _test_x_key_shows_error(self, **selection_kwargs):
        """
        Test that X key shows error with given selection.
        """
        control = self.disasm_view._flow_graph

        with (
            self._mock_infodock_selection(control, **selection_kwargs),
            patch("angrmanagement.ui.widgets.qdisasm_base_control.QMessageBox.critical") as mock_msgbox,
        ):
            control.keyPressEvent(QKeyEvent(QKeyEvent.Type.KeyPress, Qt.Key.Key_X, Qt.KeyboardModifier.NoModifier))

            mock_msgbox.assert_called_once()
            call_args = mock_msgbox.call_args
            assert "Invalid selection for XRefs" in call_args[0][1]

    def test_x_key_with_no_selection_shows_error(self):
        """Test that pressing 'x' key with no valid selection shows error message."""
        self._test_x_key_shows_error()

    def test_x_key_with_non_function_block_tree_node_shows_error(self):
        """Test that pressing 'x' key with non-function block tree node shows error message."""
        mock_block_tree_node = MagicMock()
        mock_block_tree_node.am_none = False
        mock_block_tree_node.am_obj = ("some_other_type", 0x5000)

        self._test_x_key_shows_error(block_tree_node=mock_block_tree_node)

    def test_other_key_delegates_to_base_class(self):
        """Test that pressing other keys delegates to base class keyPressEvent."""
        control = self.disasm_view._flow_graph

        with patch.object(control._base_cls, "keyPressEvent") as mock_base_key_event:
            key_event = QKeyEvent(QKeyEvent.Type.KeyPress, Qt.Key.Key_A, Qt.KeyboardModifier.NoModifier)
            control.keyPressEvent(key_event)
            mock_base_key_event.assert_called_once_with(control, key_event)


class TestContextMenus(TestDisassemblyViewBase):
    """Test context menu creation and action routing."""

    def test_instruction_context_menu_sets_address_and_creates_qmenu(self):
        """Test that instruction_context_menu sets insn_addr and creates QMenu."""
        mock_insn = MagicMock()
        mock_insn.addr = 0x1000

        assert self.disasm_view._insn_menu is not None
        with patch.object(self.disasm_view._insn_menu, "qmenu") as mock_qmenu:
            mock_qmenu.return_value = MagicMock()
            mock_qmenu.return_value.exec_ = MagicMock()

            self.disasm_view.instruction_context_menu(mock_insn, QPointF(100, 100))

            assert self.disasm_view._insn_menu.insn_addr == 0x1000
            mock_qmenu.assert_called_once()
            call_kwargs = mock_qmenu.call_args[1]
            assert "extra_entries" in call_kwargs
            assert call_kwargs.get("cached") is False

    def test_label_context_menu_sets_address_and_creates_qmenu(self):
        """Test that label_context_menu sets addr and creates QMenu."""
        assert self.disasm_view._label_menu is not None
        with patch.object(self.disasm_view._label_menu, "qmenu") as mock_qmenu:
            mock_qmenu.return_value = MagicMock()
            mock_qmenu.return_value.exec_ = MagicMock()

            self.disasm_view.label_context_menu(0x2000, QPointF(100, 100))

            assert self.disasm_view._label_menu.addr == 0x2000
            mock_qmenu.assert_called_once()

    def test_insn_menu_toggle_selection(self):
        """Test that instruction menu 'Toggle selection' calls toggle_instruction_selection."""
        assert self.disasm_view._insn_menu is not None
        self.disasm_view._insn_menu.insn_addr = 0x1000

        with patch.object(self.disasm_view.infodock, "toggle_instruction_selection") as mock_toggle:
            self.disasm_view._insn_menu._toggle_instruction_selection()
            mock_toggle.assert_called_once_with(0x1000)

    def test_insn_menu_popup_xrefs(self):
        """Test that instruction menu 'XRefs...' calls parse_operand_and_popup_xref_dialog."""
        assert self.disasm_view._insn_menu is not None
        self.disasm_view._insn_menu.insn_addr = 0x1000

        mock_operand = MagicMock()
        with (
            patch.object(
                self.disasm_view._flow_graph,
                "get_selected_operand_info",
                return_value=(MagicMock(), 0x1000, mock_operand),
            ),
            patch.object(self.disasm_view, "parse_operand_and_popup_xref_dialog") as mock_popup,
        ):
            self.disasm_view._insn_menu._popup_xrefs()
            mock_popup.assert_called_once_with(0x1000, mock_operand)

    def test_insn_menu_popup_newstate_dialog(self):
        """Test that instruction menu 'Execute symbolically...' calls popup_newstate_dialog."""
        assert self.disasm_view._insn_menu is not None
        self.disasm_view._insn_menu.insn_addr = 0x1000

        with patch.object(self.disasm_view, "popup_newstate_dialog") as mock_popup:
            self.disasm_view._insn_menu._popup_newstate_dialog()
            mock_popup.assert_called_once()

    def test_insn_menu_add_hook(self):
        """Test that instruction menu 'Add hook...' calls popup_hook_dialog."""
        assert self.disasm_view._insn_menu is not None
        self.disasm_view._insn_menu.insn_addr = 0x1000

        with patch.object(self.disasm_view, "popup_hook_dialog") as mock_popup:
            self.disasm_view._insn_menu._add_hook()
            mock_popup.assert_called_once()

    def test_insn_menu_toggle_breakpoint(self):
        """Test that instruction menu 'Toggle breakpoint' calls toggle_exec_breakpoint."""
        assert self.disasm_view._insn_menu is not None
        self.disasm_view._insn_menu.insn_addr = 0x1000

        with (
            patch.object(self.instance.breakpoint_mgr, "toggle_exec_breakpoint") as mock_toggle,
            patch.object(self.disasm_view, "refresh"),
        ):
            self.disasm_view._insn_menu._toggle_breakpoint()
            mock_toggle.assert_called_once_with(0x1000)

    def test_insn_menu_popup_patch_dialog(self):
        """Test that instruction menu 'Patch...' calls popup_patch_dialog."""
        assert self.disasm_view._insn_menu is not None
        self.disasm_view._insn_menu.insn_addr = 0x1000

        with patch.object(self.disasm_view, "popup_patch_dialog") as mock_popup:
            self.disasm_view._insn_menu._popup_patch_dialog()
            mock_popup.assert_called_once()

    def test_label_menu_popup_newstate_dialog(self):
        """Test that label menu 'Execute symbolically...' calls popup_newstate_dialog."""
        assert self.disasm_view._label_menu is not None
        self.disasm_view._label_menu.addr = 0x2000

        with patch.object(self.disasm_view, "popup_newstate_dialog") as mock_popup:
            self.disasm_view._label_menu._popup_newstate_dialog()
            mock_popup.assert_called_once()

    def test_label_menu_toggle_selection(self):
        """Test that label menu 'Toggle selection' calls toggle_label_selection."""
        assert self.disasm_view._label_menu is not None
        self.disasm_view._label_menu.addr = 0x2000

        with patch.object(self.disasm_view.infodock, "toggle_label_selection") as mock_toggle:
            self.disasm_view._label_menu._toggle_label_selection()
            mock_toggle.assert_called_once_with(0x2000)

    def test_label_menu_popup_xrefs(self):
        """Test that label menu 'XRefs...' calls popup_xref_dialog."""
        assert self.disasm_view._label_menu is not None
        self.disasm_view._label_menu.addr = 0x2000

        with patch.object(self.disasm_view, "popup_xref_dialog") as mock_popup:
            self.disasm_view._label_menu._popup_xrefs()
            mock_popup.assert_called_once_with(addr=0x2000, variable=None, dst_addr=0x2000)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from .breakpoints_view import BreakpointsView
from .call_explorer_view import CallExplorerView
from .code_view import CodeView
from .console_view import ConsoleView
from .data_dep_view import DataDepView
from .dep_view import DependencyView
from .disassembly_view import DisassemblyView
from .functions_view import FunctionsView
from .hex_view import HexView
from .jobs_view import JobsView
from .log_view import LogView
from .patches_view import PatchesView
from .proximity_view import ProximityView
from .registers_view import RegistersView
from .signatures_view import SignaturesView
from .stack_view import StackView
from .states_view import StatesView
from .strings_view import StringsView
from .symexec_view import SymexecView
from .trace_map_view import TraceMapView
from .traces_view import TracesView
from .types_view import TypesView
from .view import BaseView, InstanceView

__all__ = [
    "BreakpointsView",
    "CallExplorerView",
    "CodeView",
    "ConsoleView",
    "DataDepView",
    "DependencyView",
    "DisassemblyView",
    "FunctionsView",
    "HexView",
    "JobsView",
    "LogView",
    "PatchesView",
    "ProximityView",
    "RegistersView",
    "StackView",
    "StatesView",
    "StringsView",
    "SymexecView",
    "SignaturesView",
    "TraceMapView",
    "TracesView",
    "TypesView",
    "BaseView",
    "InstanceView",
]



from __future__ import annotations

import logging

from PySide6.QtCore import QEvent, QMarginsF, QPoint, QRectF, QSize, Qt, Signal
from PySide6.QtGui import QImage, QMouseEvent, QPainter, QVector2D
from PySide6.QtWidgets import (
    QApplication,
    QGestureEvent,
    QGraphicsScene,
    QGraphicsSceneMouseEvent,
    QGraphicsView,
    QPinchGesture,
    QStyleOptionGraphicsItem,
)

_l = logging.getLogger(__name__)


class QBaseGraphicsView(QGraphicsView):
    """QBaseGraphicsView is a QGraphicsView that emits a signal when the visible scene rect changes."""

    visible_scene_rect_changed = Signal(QRectF)

    #
    # Public methods
    #

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._visibile_scene_rect: QRectF = QRectF()

    @property
    def visible_scene_rect(self):
        return self._visibile_scene_rect

    def redraw(self) -> None:
        """
        Redraw the scene. Do not recompute any items in the view.

        :return:    None
        """
        scene = self.scene()
        if scene is not None:
            scene.update(self.sceneRect())

    def viewportEvent(self, event: QEvent) -> bool:
        visible_scene_rect = self.mapToScene(self.viewport().geometry()).boundingRect()
        if visible_scene_rect != self._visibile_scene_rect:
            self._visibile_scene_rect = visible_scene_rect
            self.visible_scene_rect_changed.emit(visible_scene_rect)

        return super().viewportEvent(event)


class QSaveableGraphicsView(QBaseGraphicsView):
    """QSaveableGraphicsView is a QGraphicsView that can save the visible scene to an image file."""

    def __init__(self, parent=None) -> None:
        super().__init__(parent=parent)
        self._is_extra_render_pass: bool = False

    @property
    def is_extra_render_pass(self):
        return self._is_extra_render_pass

    def set_extra_render_pass(self, is_extra_pass: bool) -> None:
        """
        Trigger any post-render callbacks
        """
        self._is_extra_render_pass = is_extra_pass

    def save_image_to(
        self, path, top_margin: int = 50, bottom_margin: int = 50, left_margin: int = 50, right_margin: int = 50
    ) -> None:
        margins = QMarginsF(left_margin, top_margin, right_margin, bottom_margin)

        oldRect = self.scene().sceneRect()
        minRect = self.scene().itemsBoundingRect()
        imgRect = minRect.marginsAdded(margins)

        image = QImage(imgRect.size().toSize(), QImage.Format.Format_ARGB32)
        image.fill(Qt.GlobalColor.white)
        painter = QPainter(image)

        painter.setRenderHints(QPainter.RenderHint.Antialiasing | QPainter.RenderHint.SmoothPixmapTransform)

        # draw the image
        self.scene().setSceneRect(imgRect)
        self.scene().render(painter)
        image.save(path)

        # cleanup
        painter.end()

        # restore the old scene rect
        self.scene().setSceneRect(oldRect)


class QZoomableDraggableGraphicsView(QSaveableGraphicsView):
    """QZoomableDraggableGraphicsView is a QGraphicsView that allows zooming and dragging."""

    ZOOM_X = True
    ZOOM_Y = True

    def __init__(self, parent=None) -> None:
        super().__init__(parent=parent)

        self._is_dragging = False
        self._is_mouse_pressed = False

        self._last_coords = None
        self._last_screen_pos = None

        self.setTransformationAnchor(QGraphicsView.ViewportAnchor.NoAnchor)
        self.setResizeAnchor(QGraphicsView.ViewportAnchor.AnchorViewCenter)

        # scroll bars are useless when the scene is near-infinite
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)

        # the zoom factor, for preserving the zoom
        self.zoom_factor = None

        self.grabGesture(Qt.GestureType.PinchGesture)
        self._gesture_last_scale = 1.0

    def _initial_position(self):
        raise NotImplementedError

    def _reset_view(self) -> None:
        self.resetTransform()
        self.centerOn(self._initial_position())
        self.zoom(restore=True)

    def _reset_scene(self) -> None:
        if self.scene() is None:
            scene = QGraphicsScene(self)
            self.setScene(scene)
        else:
            self.scene().clear()

    def sizeHint(self):  # pylint:disable=no-self-use
        return QSize(300, 300)

    def zoom(
        self, out: bool = False, at=None, reset: bool = False, restore: bool = False, factor: float = 1.25
    ) -> None:
        if at is None:
            at = self.scene().sceneRect().center().toPoint()
        lod = QStyleOptionGraphicsItem.levelOfDetailFromTransform(self.transform())
        zoomInFactor = factor
        zoomOutFactor = 1 / zoomInFactor

        if reset:
            zoomFactor = 1 / lod
        elif restore:
            zoomFactor = self.zoom_factor if self.zoom_factor else 1 / lod
        elif not out:
            zoomFactor = zoomInFactor
        else:
            zoomFactor = zoomOutFactor
            # limit the scroll out limit for usability
            if lod < 0.015:
                return

        # Save the scene pos
        oldPos = self.mapToScene(at)

        # Zoom
        self.scale(zoomFactor if self.ZOOM_X else 1, zoomFactor if self.ZOOM_Y else 1)
        self.zoom_factor = QStyleOptionGraphicsItem.levelOfDetailFromTransform(self.transform())

        # Get the new position
        newPos = self.mapToScene(at)

        # Move scene to old position
        delta = newPos - oldPos
        self.translate(delta.x(), delta.y())

    def wheelEvent(self, event) -> None:
        if event.modifiers() & Qt.KeyboardModifier.ControlModifier == Qt.KeyboardModifier.ControlModifier:
            self.zoom(
                event.angleDelta().y() < 0,
                QPoint(event.position().x(), event.position().y()),
                factor=1 + 0.25 * abs(event.angleDelta().y()) / 120,
            )
        elif event.angleDelta().x() != 0:
            # if this is an angled zoom (e.g. touchpad) then just let the default handler take care of it
            super().wheelEvent(event)
        else:
            # if it is not an angled zoom (e.g. mouse wheel) parse the shift key specially to mean horizontal movement
            if event.modifiers() & Qt.KeyboardModifier.ShiftModifier == Qt.KeyboardModifier.ShiftModifier:
                event.setModifiers(event.modifiers() & ~Qt.KeyboardModifier.ShiftModifier)
                self.horizontalScrollBar().wheelEvent(event)
            else:
                self.verticalScrollBar().wheelEvent(event)

    def _save_last_coords(self, event) -> None:
        pos = self.mapToScene(event.pos())
        self._last_coords = (pos.x(), pos.y())
        self._last_screen_pos = event.pos()

    def keyPressEvent(self, event) -> None:
        if event.key() == Qt.Key.Key_Equal and (
            event.modifiers() & Qt.KeyboardModifier.ControlModifier == Qt.KeyboardModifier.ControlModifier
        ):
            self.zoom(out=False)
        elif event.key() == Qt.Key.Key_Minus and (
            event.modifiers() & Qt.KeyboardModifier.ControlModifier == Qt.KeyboardModifier.ControlModifier
        ):
            self.zoom(out=True)
        elif event.key() == Qt.Key.Key_0 and (
            event.modifiers() & Qt.KeyboardModifier.ControlModifier == Qt.KeyboardModifier.ControlModifier
        ):
            self.zoom(reset=True)
        else:
            super().keyPressEvent(event)

    def mousePressEvent(self, event) -> None:
        # _l.debug('Received press')
        if event.button() == Qt.MouseButton.LeftButton:
            self._is_mouse_pressed = True
            self._is_dragging = False

            self._save_last_coords(event)
            event.accept()

        super().mousePressEvent(event)

    def mouseMoveEvent(self, event) -> None:
        """

        :param QMouseEvent event:
        :return:
        """

        SENSITIVITY = 1.0
        if self._is_mouse_pressed:
            mouse_delta = QVector2D(event.pos() - self._last_screen_pos).length()
            if mouse_delta > SENSITIVITY:
                self._is_dragging = True
                pos = self.mapToScene(event.pos())

                self.viewport().setCursor(Qt.CursorShape.ClosedHandCursor)

                delta = (pos.x() - self._last_coords[0], pos.y() - self._last_coords[1])
                self.translate(*delta)

            self._save_last_coords(event)
            event.accept()

        super().mouseMoveEvent(event)

    def dispatchMouseEventToScene(self, event: QMouseEvent):
        if event.type() == QEvent.Type.MouseButtonPress:
            newtype = QEvent.Type.GraphicsSceneMousePress
        elif event.type() == QEvent.Type.MouseButtonRelease:
            newtype = QEvent.Type.GraphicsSceneMouseRelease
        else:
            raise ValueError(f"Unknown event type {event.type()}")

        # pulled from QGraphicsView::mousePressEvent in the qt codebase:
        mouseEvent = QGraphicsSceneMouseEvent(newtype)
        mousePressViewPoint = event.pos()
        mousePressScenePoint = self.mapToScene(mousePressViewPoint)
        mousePressScreenPoint = event.globalPos()
        lastMouseMoveScenePoint = mousePressScenePoint
        lastMouseMoveScreenPoint = mousePressScreenPoint
        mousePressButton = event.button()

        # mouseEvent.setWidget(self.viewport()) # TODO figure out how to do this in python, or if it's really necessary
        mouseEvent.setButtonDownScenePos(mousePressButton, mousePressScenePoint)
        mouseEvent.setButtonDownScreenPos(mousePressButton, mousePressScreenPoint)
        mouseEvent.setScenePos(mousePressScenePoint)
        mouseEvent.setScreenPos(mousePressScreenPoint)
        mouseEvent.setLastScenePos(lastMouseMoveScenePoint)
        mouseEvent.setLastScreenPos(lastMouseMoveScreenPoint)
        mouseEvent.setButtons(event.buttons())
        mouseEvent.setButton(event.button())
        mouseEvent.setModifiers(event.modifiers())
        mouseEvent.setSource(event.source())
        mouseEvent.setFlags(event.flags())
        mouseEvent.setAccepted(False)
        QApplication.sendEvent(self.scene(), mouseEvent)
        return mouseEvent

    def mouseReleaseEvent(self, event) -> None:
        """

        :param QMouseEvent event:
        :return:
        """

        if event.button() == Qt.MouseButton.LeftButton and self._is_dragging:
            self.viewport().setCursor(Qt.CursorShape.ArrowCursor)
            event.accept()

        if not event.isAccepted():
            # create a new event and dispatch it to the scene
            pressy = QMouseEvent(
                QEvent.Type.MouseButtonPress,
                event.pos(),
                event.globalPos(),
                event.button(),
                event.buttons(),
                event.modifiers(),
            )

            _ = self.dispatchMouseEventToScene(pressy)

            releasy = QMouseEvent(
                QEvent.Type.MouseButtonRelease,
                event.pos(),
                event.globalPos(),
                event.buton(),
                event.buttons(),
                event.modifiers(),
            )
            release_event = self.dispatchMouseEventToScene(releasy)

            if not release_event.isAccepted():
                release_event.accept()

        self._is_mouse_pressed = False
        self._is_dragging = False

        super().mouseReleaseEvent(event)

    def event(self, event):
        if event.type() == QEvent.Type.Gesture and isinstance(event, QGestureEvent):
            return self._handle_pinch_gesture(event)
        return super().event(event)

    def _handle_pinch_gesture(self, event: QGestureEvent) -> None:
        for gesture in event.gestures():
            if isinstance(gesture, QPinchGesture):
                if gesture.state() == Qt.GestureState.GestureStarted:
                    self.pinchGestureStarted()
                elif gesture.state() == Qt.GestureState.GestureUpdated:
                    self.pinchGestureUpdated(gesture)

    def pinchGestureStarted(self) -> None:
        self._gesture_last_scale = 1.0

    def pinchGestureUpdated(self, gesture: QPinchGesture) -> None:
        newScale = gesture.scaleFactor() / self._gesture_last_scale
        self.zoom(at=gesture.centerPoint().toPoint(), factor=newScale)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from PySide6.QtCore import QAbstractTableModel, QSize, Qt
from PySide6.QtWidgets import QAbstractItemView, QHeaderView, QMenu, QTableView, QVBoxLayout

from angrmanagement.data.breakpoint import Breakpoint, BreakpointManager, BreakpointType
from angrmanagement.ui.dialogs import BreakpointDialog

from .view import InstanceView

if TYPE_CHECKING:
    from collections.abc import Sequence

    import PySide6

    from angrmanagement.data.instance import Instance
    from angrmanagement.ui.workspace import Workspace


class QBreakpointTableModel(QAbstractTableModel):
    """
    Breakpoint table model.
    """

    Headers = ["Type", "Address", "Size", "Comment"]
    COL_TYPE = 0
    COL_ADDR = 1
    COL_SIZE = 2
    COL_COMMENT = 3

    def __init__(self, breakpoint_mgr: BreakpointManager) -> None:
        super().__init__()
        self.breakpoint_mgr = breakpoint_mgr
        self.breakpoint_mgr.breakpoints.am_subscribe(self._on_breakpoints_updated)

    def _on_breakpoints_updated(self, **kwargs) -> None:  # pylint:disable=unused-argument
        self.beginResetModel()
        self.endResetModel()

    def rowCount(self, parent: PySide6.QtCore.QModelIndex = ...) -> int:  # pylint:disable=unused-argument
        return len(self.breakpoint_mgr.breakpoints)

    def columnCount(self, parent: PySide6.QtCore.QModelIndex = ...) -> int:  # pylint:disable=unused-argument
        return len(self.Headers)

    def headerData(self, section: int, orientation: PySide6.QtCore.Qt.Orientation, role: int = ...) -> Any:  # pylint:disable=unused-argument
        if role != Qt.ItemDataRole.DisplayRole:
            return None
        if section < len(self.Headers):
            return self.Headers[section]
        return None

    def data(self, index: PySide6.QtCore.QModelIndex, role: int = ...) -> Any:
        if not index.isValid():
            return None
        row = index.row()
        if row >= len(self.breakpoint_mgr.breakpoints):
            return None
        col = index.column()
        if role == Qt.ItemDataRole.DisplayRole:
            return self._get_column_text(self.breakpoint_mgr.breakpoints[row], col)
        else:
            return None

    def _get_column_text(self, bp: Breakpoint, column: int) -> str:
        if column == self.COL_TYPE:
            return {BreakpointType.Execute: "Execute", BreakpointType.Read: "Read", BreakpointType.Write: "Write"}.get(
                bp.type
            )
        elif column == self.COL_ADDR:
            return f"{bp.addr:#08x}"
        elif column == self.COL_SIZE:
            return f"{bp.size:#x}"
        elif column == self.COL_COMMENT:
            return bp.comment
        else:
            raise AssertionError


class QBreakpointTableWidget(QTableView):
    """
    Breakpoint table widget.
    """

    def __init__(self, breakpoint_mgr: BreakpointManager, workspace: Workspace) -> None:
        super().__init__()
        self.workspace = workspace
        self.breakpoint_mgr = breakpoint_mgr

        hheader = self.horizontalHeader()
        hheader.setVisible(True)

        vheader = self.verticalHeader()
        vheader.setVisible(False)
        vheader.setDefaultSectionSize(20)

        self.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.setHorizontalScrollMode(QAbstractItemView.ScrollMode.ScrollPerPixel)

        self.model: QBreakpointTableModel = QBreakpointTableModel(self.breakpoint_mgr)
        self.setModel(self.model)

        for col in range(len(QBreakpointTableModel.Headers)):
            hheader.setSectionResizeMode(col, QHeaderView.ResizeMode.ResizeToContents)
        hheader.setStretchLastSection(True)
        self.doubleClicked.connect(self._on_cell_double_click)

    #
    # Events
    #

    def closeEvent(self, event) -> None:
        self.model.shutdown()
        super().closeEvent(event)

    def contextMenuEvent(self, event) -> None:
        selected_rows = {i.row() for i in self.selectedIndexes()}
        breakpoints = [self.breakpoint_mgr.breakpoints[r] for r in selected_rows]
        menu = QMenu("", self)
        if len(breakpoints):
            if len(breakpoints) == 1:
                menu.addAction("Edit breakpoint", lambda: self.edit_breakpoint(breakpoints[0]))
            menu.addAction(
                "Remove breakpoint" + ("s" if len(breakpoints) > 1 else ""),
                lambda: self.remove_breakpoints(breakpoints),
            )
            menu.addSeparator()
        menu.addAction("New breakpoint", self.new_breakpoint)
        menu.exec_(event.globalPos())

    def _on_cell_double_click(self, index) -> None:
        self.edit_breakpoint(self.breakpoint_mgr.breakpoints[index.row()])

    def new_breakpoint(self) -> None:
        bp = Breakpoint(BreakpointType.Execute, 0)
        if BreakpointDialog(bp, self.workspace, self).exec_():
            self.breakpoint_mgr.add_breakpoint(bp)

    def edit_breakpoint(self, breakpoint_: Breakpoint) -> None:
        BreakpointDialog(breakpoint_, self.workspace, self).exec_()

    def remove_breakpoints(self, breakpoints: Sequence[Breakpoint]) -> None:
        for b in breakpoints:
            self.breakpoint_mgr.remove_breakpoint(b)


class BreakpointsView(InstanceView):
    """
    Breakpoints table view.
    """

    def __init__(self, workspace: Workspace, default_docking_position: str, instance: Instance) -> None:
        super().__init__("breakpoints", workspace, default_docking_position, instance)
        self.base_caption = "Breakpoints"
        self._tbl_widget: QBreakpointTableWidget | None = None
        self._init_widgets()
        self.reload()

    def reload(self) -> None:
        pass

    @staticmethod
    def minimumSizeHint():
        return QSize(200, 200)

    def _init_widgets(self) -> None:
        vlayout = QVBoxLayout()
        self._tbl_widget = QBreakpointTableWidget(self.instance.breakpoint_mgr, self.workspace)
        vlayout.addWidget(self._tbl_widget)
        self.setLayout(vlayout)



from __future__ import annotations

import os
from typing import TYPE_CHECKING

import angr
from PySide6.QtCore import Qt
from PySide6.QtWidgets import (
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QGridLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QTextEdit,
    QTreeWidget,
    QTreeWidgetItem,
    QVBoxLayout,
)

from angrmanagement.ui.dialogs.env_config import EnvConfig
from angrmanagement.ui.dialogs.fs_mount import FilesystemMount
from angrmanagement.ui.widgets import QStateComboBox
from angrmanagement.utils.namegen import NameGenerator

from .socket_config import SocketConfig

if TYPE_CHECKING:
    from angrmanagement.data.instance import Instance
    from angrmanagement.ui.workspace import Workspace


class StateMetadata(angr.SimStatePlugin):
    """
    Helper class for metadata.
    """

    def __init__(self) -> None:
        super().__init__()
        self.name = None  # the state's name
        self.base_name = None  # the name of the base state this was created from
        self.is_original = False  # is this the original instantiation of this name?
        self.is_base = False  # is this state created with nothing else as a base?

    def copy(self, _=None):
        c = StateMetadata()
        c.name = self.name
        c.base_name = self.base_name
        c.is_original = False
        c.is_base = False
        return c


StateMetadata.register_default("gui_data")


def is_option(o):
    return all(ch in "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_" for ch in o)


class NewState(QDialog):
    """
    Dialog to create a new simulation state.
    """

    def __init__(
        self,
        workspace: Workspace,
        instance: Instance,
        addr: int | None = None,
        create_simgr: bool = False,
        parent=None,
        push_to_instance: bool = True,
    ) -> None:
        super().__init__(parent)

        # initialization

        self.workspace = workspace
        self.instance = instance
        self.state = None  # output

        self._options = set()
        self._addr = addr
        self._create_simgr = create_simgr  # Shall we create a new simgr after clicking OK?

        # Shall we push the new state to instance.stats and call states.am_event(src="new", state=self.state)
        self._push_to_instance = push_to_instance

        self._editor: QTextEdit | None = None
        self._ok_button = None

        self._args: list[str] | None = None
        self._env_config: list[tuple[str, str]] = []
        self._fs_config: list[tuple[str, str]] | None = None
        self._sockets_config = None

        self.setWindowTitle("New State")

        self.main_layout = QVBoxLayout()

        self._init_widgets()

        self.setLayout(self.main_layout)

    #
    # Private methods
    #

    def _init_widgets(self) -> None:
        layout = QGridLayout()
        row = 0

        validation_failures = set()

        # name

        name_label = QLabel(self)
        name_label.setText("Name")

        name_box = QLineEdit(self)
        name_box.setText(NameGenerator.random_name())

        def handle_name(txt: str) -> None:
            nonlocal validation_failures
            key = {"name"}
            if txt and not any(s.gui_data.name == txt for s in self.instance.states):
                validation_failures -= key
            else:
                validation_failures |= key
            validation_update()

        name_box.textEdited.connect(handle_name)

        layout.addWidget(name_label, row, 0)
        layout.addWidget(name_box, row, 1)
        row += 1

        # address

        address_label = QLabel(self)
        address_label.setText("Address")

        address_box = QLineEdit(self)
        address_box.setText(hex(self.instance.project.entry) if self._addr is None else hex(self._addr))

        def handle_address(_) -> None:
            nonlocal validation_failures
            key = {"addr"}
            if parse_address() is not None:
                validation_failures -= key
            else:
                validation_failures |= key
            validation_update()

        def parse_address():
            txt = address_box.text()
            try:
                return self.instance.project.kb.labels.lookup(txt)
            except KeyError:
                pass

            try:
                return int(txt, 16)
            except ValueError:
                return None

        address_box.textEdited.connect(handle_address)
        layout.addWidget(address_label, row, 0)
        layout.addWidget(address_box, row, 1)
        row += 1

        # template

        template_label = QLabel(self)
        template_label.setText("Template")

        template_combo = QComboBox()
        template_combo.addItem("Blank State", "blank")
        template_combo.addItem("Call state", "call")
        template_combo.addItem("Entry state", "entry")
        template_combo.addItem("Full-init state", "full")

        def handle_template(_) -> None:
            base_allowed = template_combo.currentData() in ("blank", "call")
            base_state_combo.setHidden(not base_allowed)
            base_state_label.setHidden(not base_allowed)
            args_allowed = template_combo.currentData() in ("entry",)
            args_label.setHidden(not args_allowed)
            args_edit.setHidden(not args_allowed)
            env_label.setHidden(not args_allowed)
            env_button.setHidden(not args_allowed)

        template_combo.currentIndexChanged.connect(handle_template)

        layout.addWidget(template_label, row, 0)
        layout.addWidget(template_combo, row, 1)
        row += 1

        # base state

        base_state_label = QLabel(self)
        base_state_label.setText("Base state")

        base_state_combo = QStateComboBox(self.instance, self)

        layout.addWidget(base_state_label, row, 0)
        layout.addWidget(base_state_combo, row, 1)
        row += 1

        # args
        args_label = QLabel(self)
        args_label.setText("Args")

        args_edit = QTextEdit(self)
        args_edit.setAcceptRichText(False)
        args_edit.setFixedHeight(60)

        def handle_args() -> None:
            self._args = [self.instance.project.filename.encode() or b"dummy_filename"] + [
                x.encode() for x in args_edit.toPlainText().split()
            ]

        args_edit.textChanged.connect(handle_args)

        layout.addWidget(args_label, row, 0)
        layout.addWidget(args_edit, row, 1)
        row += 1

        # env_config
        env_label = QLabel(self)
        env_label.setText("Environment")
        env_button = QPushButton(self)
        env_button.setText("Change")

        layout.addWidget(env_label, row, 0)
        layout.addWidget(env_button, row, 1)

        def env_edit_button() -> None:
            env_dialog = EnvConfig(env_config=self._env_config, instance=self.instance, parent=self)
            env_dialog.exec_()
            self._env_config = env_dialog.env_config
            env_button.setText(f"{len(self._env_config)} Items")

        env_button.clicked.connect(env_edit_button)

        row += 1

        # fs_mount
        fs_label = QLabel(self)
        fs_label.setText("Filesystem")
        fs_button = QPushButton(self)
        fs_button.setText("Change")

        layout.addWidget(fs_label, row, 0)
        layout.addWidget(fs_button, row, 1)

        def fs_edit_button() -> None:
            fs_dialog = FilesystemMount(fs_config=self._fs_config, instance=self.instance, parent=self)
            fs_dialog.exec_()
            self._fs_config = fs_dialog.fs_config
            fs_button.setText(f"{len(self._fs_config)} Items")

        fs_button.clicked.connect(fs_edit_button)

        row += 1

        # socket support
        socket_label = QLabel(self)
        socket_label.setText("Sockets")
        socket_button = QPushButton(self)
        socket_button.setText("Change")

        layout.addWidget(socket_label, row, 0)
        layout.addWidget(socket_button, row, 1)

        def socket_edit_button() -> None:
            socket_dialog = SocketConfig(socket_config=self._sockets_config, instance=self.instance)
            socket_dialog.exec_()
            self._sockets_config = socket_dialog.socket_config
            # socket_button.setText("{} Items".format(len(self._sockets_config)))

        socket_button.clicked.connect(socket_edit_button)

        row += 1

        # mode

        mode_label = QLabel(self)
        mode_label.setText("Mode")

        mode_combo = QComboBox(self)
        mode_combo.addItem("Symbolic", "symbolic")
        mode_combo.addItem("Static", "static")
        mode_combo.addItem("Fast-path", "fastpath")
        mode_combo.addItem("Tracing", "tracing")

        def mode_changed() -> None:
            self._options.clear()
            self._options.update(angr.sim_options.modes[mode_combo.currentData()])
            for child in children_items:
                child.setCheckState(
                    0, Qt.CheckState.Checked if child.text(0) in self._options else Qt.CheckState.Unchecked
                )

        mode_combo.currentIndexChanged.connect(mode_changed)
        self._options.clear()
        self._options.update(angr.sim_options.modes[mode_combo.currentData()])

        layout.addWidget(mode_label, row, 0)
        layout.addWidget(mode_combo, row, 1)
        row += 1

        # options tree

        options_label = QLabel(self)
        options_label.setText("Options")

        options_tree = QTreeWidget(self)
        options_tree.setHeaderHidden(True)
        children_items = []
        for name, members in angr.sim_options.__dict__.items():
            if not isinstance(members, set):
                continue
            if name == "resilience_options":
                continue
            parent = QTreeWidgetItem(options_tree)
            parent.setText(0, name)
            parent.setFlags(parent.flags() | Qt.ItemFlag.ItemIsAutoTristate | Qt.ItemFlag.ItemIsUserCheckable)
            for option in members:
                child = QTreeWidgetItem(parent)
                child.setText(0, option)
                child.setFlags(child.flags() | Qt.ItemFlag.ItemIsUserCheckable)
                child.setCheckState(0, Qt.CheckState.Checked if option in self._options else Qt.CheckState.Unchecked)
                children_items.append(child)
        parent = QTreeWidgetItem(options_tree)
        parent.setText(0, "All options")
        parent.setFlags(parent.flags() | Qt.ItemFlag.ItemIsAutoTristate | Qt.ItemFlag.ItemIsUserCheckable)
        for option in {x for x in angr.sim_options.__dict__.values() if isinstance(x, str) and is_option(x)}:
            child = QTreeWidgetItem(parent)
            child.setText(0, option)
            child.setFlags(child.flags() | Qt.ItemFlag.ItemIsUserCheckable)
            child.setCheckState(0, Qt.CheckState.Checked if option in self._options else Qt.CheckState.Unchecked)
            children_items.append(child)

        def maintain_model(item: QTreeWidgetItem, _) -> None:
            option = item.text(0)
            if not is_option(option):
                return

            checked = item.checkState(0) == Qt.CheckState.Checked
            if (option in self._options) == checked:
                return

            if checked:
                self._options.add(option)
            else:
                self._options.remove(option)

            for child in children_items:
                if child is not item and child.text(0) == option:
                    child.setCheckState(0, Qt.CheckState.Checked if checked else Qt.CheckState.Unchecked)

        options_tree.itemChanged.connect(maintain_model)

        layout.addWidget(options_label, row, 0)
        layout.addWidget(options_tree, row, 1)
        row += 1

        # options filter

        options_filter_label = QLabel(self)
        options_filter_label.setText("")

        options_filter_box = QLineEdit(self)
        options_filter_box.setPlaceholderText("Filter")

        def do_filter(text: str) -> None:
            for child in children_items:
                child.setHidden(text.upper() not in child.text(0))

        options_filter_box.textEdited.connect(do_filter)

        layout.addWidget(options_filter_label, row, 0)
        layout.addWidget(options_filter_box, row, 1)
        row += 1

        # buttons

        buttons = QDialogButtonBox(parent=self)
        buttons.setStandardButtons(QDialogButtonBox.StandardButton.Cancel | QDialogButtonBox.StandardButton.Ok)

        def do_ok() -> None:
            name = name_box.text()
            template = template_combo.currentData()
            addr = parse_address()
            base_state = base_state_combo.state
            mode = mode_combo.currentData()
            factory = self.instance.project.factory
            if template in ("blank", "call") and base_state is not None:
                if template == "blank":
                    self.state = factory.blank_state(addr=addr, base_state=base_state, options=self._options)
                else:
                    self.state = factory.call_state(addr, base_state=base_state, options=self._options)
                self.state.gui_data.base_name = base_state.gui_data.name
            else:
                if template == "blank":
                    self.state = factory.blank_state(addr=addr, mode=mode, options=self._options)
                elif template == "call":
                    self.state = factory.call_state(addr, mode=mode, options=self._options)
                elif template == "entry":
                    self.state = factory.entry_state(
                        mode=mode, options=self._options, args=self._args, env=dict(self._env_config)
                    )
                else:
                    self.state = factory.full_init_state(mode=mode, options=self._options)
                self.state.gui_data.base_name = name
                self.state.gui_data.is_base = True

            self.state.gui_data.name = name
            self.state.gui_data.is_original = True

            # mount fs
            if self._fs_config:
                for path, real in self._fs_config:
                    if os.path.isdir(real):
                        fs = angr.SimHostFilesystem(real)
                        fs.set_state(self.state)
                        self.state.fs.mount(path, fs)

            if self._sockets_config:
                self.state.posix.sockets = self._sockets_config.convert()

            if self._push_to_instance:
                states_list = self.instance.states
                states_list.append(self.state)
                states_list.am_event(src="new", state=self.state)

            if self._create_simgr:
                self.workspace.create_simulation_manager(self.state, name)

            self.close()

        ok_button = buttons.button(QDialogButtonBox.StandardButton.Ok)
        buttons.accepted.connect(do_ok)

        def validation_update() -> None:
            ok_button.setDisabled(bool(validation_failures))

        buttons.rejected.connect(self.close)

        self.main_layout.addLayout(layout)
        self.main_layout.addWidget(buttons)

        handle_template(None)



from __future__ import annotations

from typing import TYPE_CHECKING

from PySide6.QtCore import QRectF, Qt
from PySide6.QtGui import QPainter
from PySide6.QtWidgets import QGraphicsSimpleTextItem

from angrmanagement.config import Conf

from .qgraph_object import QCachedGraphicsItem

if TYPE_CHECKING:
    from angrmanagement.data.instance import Instance


class QBlockLabel(QCachedGraphicsItem):
    def __init__(self, addr: int, text: str, config, disasm_view, instance: Instance, infodock, parent=None) -> None:
        super().__init__(parent=parent)

        self.instance = instance
        self.addr = addr
        self.text = text
        self._width = 0
        self._height = 0
        self.infodock = infodock

        self._config = config
        self._disasm_view = disasm_view

        self._text_item: QGraphicsSimpleTextItem = None

        self._init_widgets()

    def setVisible(self, visible):
        super().setVisible(visible)
        self._text_item.setVisible(visible)

    def paint(self, painter, option, widget) -> None:  # pylint: disable=unused-argument
        painter.setRenderHints(QPainter.RenderHint.Antialiasing | QPainter.RenderHint.SmoothPixmapTransform)
        painter.setFont(self._config.code_font)

        # background
        if self.infodock.is_label_selected(self.addr):
            highlight_color = Conf.disasm_view_label_highlight_color
            painter.setBrush(highlight_color)
            painter.setPen(highlight_color)
            painter.drawRect(0, 0, self.width, self.height)

    #
    # Event handlers
    #

    def mousePressEvent(self, event) -> None:
        if event.button() == Qt.MouseButton.LeftButton:
            self.infodock.select_label(self.addr)
        elif event.button() == Qt.MouseButton.RightButton:
            # unselect all other labels
            self.infodock.unselect_all_labels()
            # select this label
            self.infodock.select_label(self.addr)
            # show context menu
            self.infodock.disasm_view.label_context_menu(self.addr, event.screenPos())

    #
    # Private methods
    #

    def _init_widgets(self) -> None:
        self._text_item = QGraphicsSimpleTextItem(self.text, self)
        self._text_item.setBrush(Conf.disasm_view_label_color)
        self._text_item.setFont(self._config.disasm_font)

        self._layout_items_and_update_size()

    def _layout_items_and_update_size(self) -> None:
        self._text_item.setPos(0, 0)

        self._width = self._text_item.boundingRect().width()
        self._height = self._text_item.boundingRect().height()
        self.recalculate_size()

    def _boundingRect(self):
        return QRectF(0, 0, self._width, self._height)



# pylint: disable=missing-class-docstring
from __future__ import annotations

import unittest
from typing import TYPE_CHECKING
from unittest.mock import patch

from common import ProjectOpenTestCase
from PySide6.QtWidgets import QApplication

from angrmanagement.ui.views import types_view

if TYPE_CHECKING:
    from angrmanagement.ui.dialogs.type_editor import CTypeEditor


class TestTypeEditor(ProjectOpenTestCase):
    def setUp(self):
        super().setUp()

        self2 = self

        def editor_exec(self, *args, **kwargs):
            assert self2._next_exec is not None
            self2._next_exec(self, *args, **kwargs)

        mock_exec = patch("PySide6.QtWidgets.QDialog.exec_", editor_exec)
        self.addCleanup(mock_exec.stop)
        mock_exec.start()

        self.types_view = types_view.TypesView(self.main.workspace, "center", self.main.workspace.main_instance)
        self.types_view.show()
        self._next_exec = None

    def editor_input(self, s):
        def wrapper(f):
            @self.next_exec
            def _(editor: CTypeEditor):
                editor.setText(s)
                editor._evaluate()

            f()
            QApplication.processEvents()

        return wrapper

    def next_exec(self, f):
        self._next_exec = f
        return f

    def tearDown(self):
        self.types_view.hide()
        del self.types_view
        super().tearDown()

    def test_wawa(self):
        QApplication.processEvents()
        assert len(self.types_view.typedefs) == 0

        @self.editor_input("struct wawa{};")
        def _a():
            self.types_view._on_new_type()

        assert len(self.types_view.typedefs) == 1
        assert set(self.project.kb.types.iter_own_keys()) == {"wawa", "struct wawa"}

        @self.editor_input("struct foo { struct wawa a; wawa b; int c; };")
        def _b():
            self.types_view._on_new_type()

        assert len(self.types_view.typedefs) == 2
        assert set(self.project.kb.types.iter_own_keys()) == {"wawa", "struct wawa", "foo", "struct foo"}
        assert self.types_view.typedefs[0].type.name == "wawa"
        assert self.types_view.typedefs[1].type.name == "foo"
        assert self.project.kb.types["wawa"].type.size == 0
        assert self.project.kb.types["foo"].type.size == 32

        @self.editor_input("struct wawa { int hey; };")
        def _c():
            self.types_view.typedefs[0].highlight = 0
            self.types_view.typedefs[0].mouseDoubleClickEvent(None)

        assert len(self.types_view.typedefs) == 2
        assert set(self.project.kb.types.iter_own_keys()) == {"wawa", "struct wawa", "foo", "struct foo"}
        assert self.types_view.typedefs[0].type.name == "wawa"
        assert self.types_view.typedefs[1].type.name == "foo"
        assert self.project.kb.types["wawa"].type.size == 32
        assert self.project.kb.types["foo"].type.size == 96


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from angr.calling_conventions import SimRegArg
from PySide6.QtGui import QTextCharFormat

from angrmanagement.config import Conf
from angrmanagement.utils.func import type2str

from .base_objects import BlockTreeNode

if TYPE_CHECKING:
    from angr.sim_type import SimTypeFunction


class QFunctionHeaderFuncNameItem(BlockTreeNode):
    """
    The function name in a function header.
    """

    def __init__(self, addr: int, name: str, infodock, parent=None):
        self.addr = addr
        self.name = name
        super().__init__(self, infodock, parent)

    def create_subobjs(self, obj) -> None:
        self.add_text(self.name)

    @property
    def selection_key(self) -> Any:
        return "func_name", self.addr

    @staticmethod
    def fmt() -> QTextCharFormat:
        fmt = QTextCharFormat()
        fmt.setFont(Conf.disasm_font)
        fmt.setForeground(Conf.disasm_view_function_color)
        return fmt


class QFunctionHeaderFuncTypeItem(BlockTreeNode):
    """
    Return type and argument types in a function header.
    """

    def __init__(self, addr: int, arg_id: int, type_str: str, infodock, parent=None):
        self.addr = addr
        self.arg_id = arg_id
        self.type_str = type_str
        super().__init__(self, infodock, parent)

    def create_subobjs(self, obj) -> None:
        self.add_text(self.type_str)

    @property
    def selection_key(self) -> Any:
        return ("param_type", self.addr, self.arg_id) if self.arg_id >= 0 else ("return_type", 0)

    @staticmethod
    def fmt() -> QTextCharFormat:
        fmt = QTextCharFormat()
        fmt.setFont(Conf.disasm_font)
        fmt.setForeground(Conf.disasm_view_function_arg_type_color)
        return fmt


class QFunctionHeaderFuncParamItem(BlockTreeNode):
    """
    Function parameters in a function header.
    """

    def __init__(self, addr: int, arg_id: int, param_str: str, infodock, parent=None):
        self.addr = addr
        self.arg_id = arg_id
        self.param_str = param_str
        super().__init__(self, infodock, parent)

    def create_subobjs(self, obj) -> None:
        self.add_text(self.param_str)

    @property
    def selection_key(self) -> Any:
        return "param_item", self.addr, self.arg_id

    @staticmethod
    def fmt() -> QTextCharFormat:
        fmt = QTextCharFormat()
        fmt.setFont(Conf.disasm_font)
        fmt.setForeground(Conf.disasm_view_function_arg_name_color)
        return fmt


class QFunctionHeaderFuncArgItem(BlockTreeNode):
    """
    Raw function arguments (in terms of registers and stack slots) in a function header.
    """

    def __init__(self, addr: int, arg_id: int, arg_str: str, infodock, parent=None):
        self.addr = addr
        self.arg_id = arg_id
        self.arg_str = arg_str
        super().__init__(self, infodock, parent)

    def create_subobjs(self, obj) -> None:
        self.add_text(self.arg_str)

    @property
    def selection_key(self) -> Any:
        return "arg_item", self.addr, self.arg_id

    @staticmethod
    def fmt() -> QTextCharFormat:
        fmt = QTextCharFormat()
        fmt.setFont(Conf.disasm_font)
        fmt.setForeground(Conf.disasm_view_function_arg_name_color)
        return fmt


class QFunctionHeader(BlockTreeNode):
    """
    Function header item in the disassembly view.

    A function header item includes the following items:
    - Function name, or demangled function name if available
    - Original function name if there is a demangled function name available
    - Function prototype: return type, argument type, and arguments
    - Calling convention
    - A list of arguments in terms of registers and memory allocations
    - Number of callers
    - Number of callees
    """

    def __init__(
        self,
        addr: int,
        name: str,
        demangled_name: str | None,
        prototype,
        args,
        infodock,
        parent=None,
        options=None,
    ) -> None:
        self.addr = addr
        self.name = name
        self.demangled_name = demangled_name
        self.prototype: SimTypeFunction = prototype
        self.args = args

        self._arg_str_list = None

        self._init_items()

        super().__init__(
            None, infodock, parent, options=options, display_address=False, top_margin_lines=2, bottom_margin_lines=0
        )

    def create_subobjs(self, obj) -> None:
        #
        # function name
        #

        if self.demangled_name and self.demangled_name != self.name:
            sub_obj = QFunctionHeaderFuncNameItem(self.addr, self.demangled_name, self.infodock, self)
            self._add_subobj(sub_obj)
            self.add_newline()

        #
        # Function prototype (including function name)
        #

        if self.prototype is None:
            # function name
            sub_obj = QFunctionHeaderFuncNameItem(self.addr, self.name, self.infodock, self)
            self._add_subobj(sub_obj)
            self.add_text("()")

        else:
            # return type
            rt = type2str(self.prototype.returnty)
            sub_obj = QFunctionHeaderFuncTypeItem(self.addr, -1, rt, self.infodock, self)
            self._add_subobj(sub_obj)
            self.add_text(" ")

            # function name
            sub_obj = QFunctionHeaderFuncNameItem(self.addr, self.name, self.infodock, self)
            self._add_subobj(sub_obj)

            self.add_text("(")

            # parameters
            for i, arg_type in enumerate(self.prototype.args):
                type_str = type2str(arg_type)
                sub_obj = QFunctionHeaderFuncTypeItem(self.addr, i, type_str, self.infodock, self)
                self._add_subobj(sub_obj)
                self.add_text(" ")

                if self.prototype.arg_names and i < len(self.prototype.arg_names):
                    param_name = self.prototype.arg_names[i]
                else:
                    param_name = f"arg_{i}"
                sub_obj = QFunctionHeaderFuncParamItem(self.addr, i, param_name, self.infodock, self)
                self._add_subobj(sub_obj)

                if i < len(self.prototype.args) - 1:
                    self.add_text(", ")

            self.add_text(")")
        self.add_newline()

        # arguments
        if self._arg_str_list is not None and len(self._arg_str_list) > 0:
            self.add_newline()
            for i, arg_str in enumerate(self._arg_str_list):
                sub_obj = QFunctionHeaderFuncArgItem(self.addr, i, arg_str, self.infodock, self)
                self.add_text(f"arg_{i} @ ")
                self._add_subobj(sub_obj)
                self.add_newline()

    def should_highlight(self) -> bool:
        #  we don't highlight the entire function header
        return False

    #
    # Private methods
    #

    def _init_items(self) -> None:
        if self.args is not None:
            self._arg_str_list = []
            for arg in self.args:
                if isinstance(arg, SimRegArg):
                    self._arg_str_list.append(arg.reg_name)
                else:
                    self._arg_str_list.append(str(arg))



from __future__ import annotations

from typing import TYPE_CHECKING

from .job import InstanceJob

if TYPE_CHECKING:
    from angrmanagement.data.instance import Instance
    from angrmanagement.logic.jobmanager import JobContext


class PrototypeFindingJob(InstanceJob):
    def __init__(self, instance: Instance, on_finish=None) -> None:
        super().__init__("Function prototype finding", instance, on_finish=on_finish)

    def run(self, ctx: JobContext) -> None:
        func_count = len(self.instance.kb.functions)
        for i, func in enumerate(self.instance.kb.functions.values()):
            if func.is_simprocedure or func.is_plt:
                func.find_declaration()

            percentage = i / func_count * 100
            ctx.set_progress(percentage)

    def __repr__(self) -> str:
        return "PrototypeFindingJob"



from __future__ import annotations

import functools
import logging
from enum import Enum
from typing import TYPE_CHECKING

import angr

from angrmanagement.data.analysis_options import (
    AnalysisConfiguration,
    BoolAnalysisOption,
    ChoiceAnalysisOption,
    StringAnalysisOption,
    extract_first_paragraph_from_docstring,
)
from angrmanagement.logic.threads import gui_thread_schedule_async

from .job import InstanceJob

if TYPE_CHECKING:
    from angrmanagement.data.instance import Instance
    from angrmanagement.logic.jobmanager import JobContext

_l = logging.getLogger(name=__name__)


class CFGForceScanMode(Enum):
    """
    CFG scanning mode options.
    """

    Disabled = 0
    SmartScan = 1
    CompleteScan = 2


class CFGAnalysisConfiguration(AnalysisConfiguration):
    """
    Configuration for CFGFast analysis.
    """

    def __init__(self, instance: Instance) -> None:
        super().__init__(instance)
        self.name = "cfg"
        self.display_name = "Control-Flow Graph Recovery"
        doc = angr.analyses.cfg.CFGFast.__doc__
        self.description = extract_first_paragraph_from_docstring(doc) if doc else ""
        self.enabled = True
        self.options = {
            o.name: o
            for o in [
                BoolAnalysisOption("resolve_indirect_jumps", "Resolve indirect jumps", True),
                BoolAnalysisOption("data_references", "Collect cross-references and guess data types", True),
                BoolAnalysisOption("cross_references", "Perform deep analysis on cross-references (slow)"),
                BoolAnalysisOption("skip_unmapped_addrs", "Skip unmapped addresses", True),
                BoolAnalysisOption("exclude_sparse_regions", "Exclude Sparse Regions", True),
                BoolAnalysisOption(
                    "explicit_analysis_starts", "Exclude non-explicit functions for analysis (incomplete)", False
                ),
                ChoiceAnalysisOption(
                    "scanning_mode",
                    "Scan to maximize identified code blocks",
                    {
                        CFGForceScanMode.Disabled: "Disabled",
                        CFGForceScanMode.SmartScan: "Smart Scan",
                        CFGForceScanMode.CompleteScan: "Complete Scan",
                    },
                    CFGForceScanMode.SmartScan,
                ),
                ChoiceAnalysisOption(
                    "function_prologues",
                    "Scan for function prologues",
                    {
                        None: "Auto",
                        True: "Enabled",
                        False: "Disabled",
                    },
                    None,
                ),
                StringAnalysisOption(
                    "regions",
                    "Regions for analysis",
                    tooltip="Specify ranges of regions for which to recover CFG. Example: 0x400000-0x401000. You may "
                    "specify multiple address ranges for analysis.",
                ),
                StringAnalysisOption(
                    "function_starts",
                    "Start at function addresses",
                    tooltip="Specify function addresses to start recursive descent of CFG generation to speed up "
                    "analysis. Example: 0x400000,0x401000",
                ),
            ]
        }

    def to_dict(self) -> dict:
        cfg_options = super().to_dict()

        # update function start locations
        if "function_starts" in cfg_options:
            function_starts = []
            for func_start_str in cfg_options["function_starts"].split(","):
                func_start_str = func_start_str.strip(" ")
                if not func_start_str:
                    continue

                try:
                    func_addr = int(func_start_str, 16)
                except ValueError as e:
                    _l.error("Invalid analysis start: %s", func_start_str)
                    raise ValueError("Invalid function start string") from e

                function_starts.append(func_addr)

            if function_starts:
                if "explicit_analysis_starts" in cfg_options:
                    cfg_options["elf_eh_frame"] = False
                    cfg_options["symbols"] = False
                    cfg_options["start_at_entry"] = False

                cfg_options["function_starts"] = function_starts

        # discard "explicit_analysis_starts" even if function_starts is not set
        if "explicit_analysis_starts" in cfg_options:
            del cfg_options["explicit_analysis_starts"]

        # update options for region specification
        if "regions" in cfg_options:
            regions = []
            for region_str in cfg_options["regions"].split(","):
                region_str = region_str.strip(" ")
                if not region_str:
                    continue
                if "-" not in region_str or region_str.count("-") != 1:
                    _l.error("Invalid analysis region: %s", region_str)
                    raise ValueError("Invalid analysis region")
                min_addr, max_addr = region_str.split("-")
                try:
                    min_addr = int(min_addr, 16)
                    max_addr = int(max_addr, 16)
                except ValueError as e:
                    _l.error("Invalid analysis region: %s", region_str)
                    raise ValueError("Invalid analysis region bound") from e
                regions.append((min_addr, max_addr))
            if regions:
                cfg_options["regions"] = regions

        scanning_mode = cfg_options.pop("scanning_mode", None)
        if scanning_mode is not None:
            cfg_options["force_smart_scan"] = scanning_mode == CFGForceScanMode.SmartScan
            cfg_options["force_complete_scan"] = scanning_mode == CFGForceScanMode.CompleteScan

        return cfg_options


class CFGGenerationJob(InstanceJob):
    """
    Job for generating the Control Flow Graph.
    """

    DEFAULT_CFG_ARGS = {
        "normalize": True,  # this is what people naturally expect
    }

    def __init__(self, instance: Instance, on_finish=None, **kwargs) -> None:
        super().__init__("CFG generation", instance, on_finish=on_finish)

        # TODO: sanitize arguments

        # make a copy
        cfg_args = dict(kwargs)
        for key, val in self.DEFAULT_CFG_ARGS.items():
            if key not in cfg_args:
                cfg_args[key] = val

        self.cfg_args = cfg_args
        self._cfb = None

    def run(self, ctx: JobContext):
        exclude_region_types = {"kernel", "tls"}
        # create a temporary CFB for displaying partially analyzed binary during CFG recovery
        temp_cfb = self.instance.project.analyses.CFB(
            exclude_region_types=exclude_region_types, on_object_added=self._on_cfb_object_added
        )
        self._cfb = temp_cfb

        cfg = self.instance.project.analyses.CFG(
            progress_callback=functools.partial(self._progress_callback, ctx),
            low_priority=True,
            cfb=temp_cfb,
            **self.cfg_args,
        )
        self._cfb = None
        # Build the real one
        cfb = self.instance.project.analyses.CFB(kb=cfg.kb, exclude_region_types=exclude_region_types)

        return cfg.model, cfb

    def __repr__(self) -> str:
        return "Generating CFG"

    #
    # Private methods
    #

    def _progress_callback(self, ctx: JobContext, percentage, text: str | None = None, cfg=None) -> None:
        ctx.set_progress(percentage, text)

        if cfg is not None:
            # Peek into the CFG
            gui_thread_schedule_async(
                self._refresh,
                args=(
                    cfg.model,
                    self._cfb,
                ),
            )

    def _refresh(self, cfg_model, cfb) -> None:
        # do not signal events. that will happen on a timer to not overwhelm the renderer
        # instance will exist because _run must be used first
        self.instance.cfg = cfg_model
        self.instance.cfb = cfb

    def _on_cfb_object_added(self, addr: int, obj) -> None:
        self.instance.cfb.am_event(object_added=(addr, obj))



#
# POPULAR CONSTANTS
#
from __future__ import annotations

import math

# Error computation
MACHINE_EPSILON_DOUBLE_PRECISION = 2.2204460493e-16
MACHINE_EPSILON_SINGLE_PRECISION = 1.1920928955e-07

# Math
PI = math.pi
E = math.e

# Physics
SPEED_OF_LIGHT = 299792458
PLANCK_CONSTANT = 6.62607004e-34
GRAVITATIONAL_CONSTANT = 6.67408e-11
ELEMENTARY_CHARGE = 1.6021766208e-19
BOLTZMANN_CONSTANT = 1.38064852e-23
AVOGADRO_CONSTANT = 6.022140857e23

# Computation
MAX_INT32 = 2147483647
MAX_INT64 = 9223372036854775807
MAX_UINT32 = 4294967295
MAX_UINT64 = 18446744073709551615
MAX_FLOAT32 = 3.40282347e38
MAX_FLOAT64 = 1.7976931348623157e308

# All consts
CONSTANTS_BY_NAME = {
    "Machine Epsilon (Double Precision)": MACHINE_EPSILON_DOUBLE_PRECISION,
    "Machine Epsilon (Single Precision)": MACHINE_EPSILON_SINGLE_PRECISION,
    "Planck Constant": PLANCK_CONSTANT,
    "Gravitational Constant": GRAVITATIONAL_CONSTANT,
    "Avogadro Constant": AVOGADRO_CONSTANT,
    "Speed of Light": SPEED_OF_LIGHT,
    "Pi": PI,
    "E": E,
}
